#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from zun.db import api as dbapi
from zun.objects import base


@base.ZunObjectRegistry.register
class QuotaClass(base.ZunPersistentObject, base.ZunObject):
    # Version 1.0: Initial version
    # Version 1.1: Add uuid column
    # Version 1.2: Remove uuid column
    VERSION = '1.2'

    fields = {
        'id': fields.IntegerField(),
        'class_name': fields.StringField(nullable=True),
        'resource': fields.StringField(nullable=True),
        'hard_limit': fields.IntegerField(nullable=True),
    }

    @staticmethod
    def _from_db_method(quota_class, db_quota_class):
        """Convert a database entity to a format object"""
        for field in quota_class.fields:
            if field == 'uuid':
                continue
            setattr(quota_class, field, db_quota_class[field])

        quota_class.obj_reset_changes()
        return quota_class

    @base.remotable_classmethod
    def get(cls, context, class_name, resource):
        """Find a quota class based on class_name and resource name.

        :param class_name: the name of class.
        :param context: security context.
        :param resource: the name of resource.
        :returns: a :class:`QuotaClass` object.
        """
        db_quota_class = dbapi.quota_class_get(context, class_name, resource)
        quota_class = QuotaClass._from_db_method(cls(context), db_quota_class)
        return quota_class

    @base.remotable_classmethod
    def get_all(cls, context, class_name=None):
        """Find quota based on class_name

        :param context: security context.
        :param class_name: the class name.
        :returns: a dict
        """
        if class_name is None:
            res = dbapi.quota_class_get_default(context)
        else:
            res = dbapi.quota_class_get_all_by_name(context, class_name)
        return res

    @base.remotable
    def create(self, context):
        """Create a QuotaClass record in the DB.

        :param context: security context. NOTE: This should only be
                        used internally by the indirection api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: QuotaClass(context)
        """
        values = self.obj_get_changes()
        class_name = values.get('class_name')
        resource = values.get('resource')
        limit = values.get('hard_limit')
        dbapi.quota_class_create(context, class_name, resource, limit)

    @base.remotable
    def update(self, context=None):
        """Save updates to this QuotaClass.

        Updates will be made column by column based on the result
        of self.what_changed().

        :param context: security context. NOTE: This should only be
                        used internally by the indirection api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: QuotaClass(context)
        """
        updates = self.obj_get_changes()
        limit = updates.get('hard_limit')
        dbapi.quota_class_update(context, self.class_name,
                                 self.resource, limit)
        self.obj_reset_changes()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import errno

import os_vif
from oslo_config import cfg
from oslo_log import log as logging
import pyroute2
from pyroute2 import netns as pyroute_netns
from stevedore import driver as stv_driver

from zun.cni import utils
from zun.common import consts
from zun.common import privileged


_BINDING_NAMESPACE = 'zun.cni.binding'
LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class BaseBindingDriver(object, metaclass=abc.ABCMeta):
    """Interface to attach ports to capsules/containers."""

    @abc.abstractmethod
    def connect(self, vif, ifname, netns, container_id):
        raise NotImplementedError()

    @abc.abstractmethod
    def disconnect(self, vif, ifname, netns, container_id):
        raise NotImplementedError()


def _get_binding_driver(vif):
    mgr = stv_driver.DriverManager(namespace=_BINDING_NAMESPACE,
                                   name=type(vif).__name__,
                                   invoke_on_load=True)
    return mgr.driver


def get_ipdb(netns=None):
    if netns:
        ipdb = pyroute2.IPDB(nl=pyroute2.NetNS(netns))
    else:
        ipdb = pyroute2.IPDB()
    return ipdb


def _enable_ipv6(netns):
    # Docker disables IPv6 for --net=none containers
    # TODO(apuimedo) remove when it is no longer the case
    try:
        path = utils.convert_netns('/proc/self/ns/net')
        self_ns_fd = open(path)
        pyroute_netns.setns(netns)
        path = utils.convert_netns('/proc/sys/net/ipv6/conf/all/disable_ipv6')
        with open(path, 'w') as disable_ipv6:
            disable_ipv6.write('0')
    except Exception:
        raise
    finally:
        pyroute_netns.setns(self_ns_fd)


@privileged.cni.entrypoint
def _configure_l3(vif_dict, ifname, netns, is_default_gateway):
    with get_ipdb(netns) as ipdb:
        with ipdb.interfaces[ifname] as iface:
            for subnet in vif_dict['network']['subnets']:
                if subnet['cidr']['version'] == 6:
                    _enable_ipv6(netns)
                for fip in subnet['ips']:
                    iface.add_ip('%s/%s' % (fip['address'],
                                            subnet['cidr']['prefixlen']))

        routes = ipdb.routes
        for subnet in vif_dict['network']['subnets']:
            for route in subnet['routes']:
                routes.add(gateway=str(route['gateway']),
                           dst=str(route['cidr'])).commit()
            if is_default_gateway and 'gateway' in subnet:
                try:
                    routes.add(gateway=str(subnet['gateway']),
                               dst='default').commit()
                except pyroute2.NetlinkError as ex:
                    if ex.code != errno.EEXIST:
                        raise
                    LOG.debug("Default route already exists in "
                              "capsule/container for vif=%s. Did not "
                              "overwrite with requested gateway=%s",
                              vif_dict, subnet['gateway'])


def _need_configure_l3(vif):
    if not hasattr(vif, 'physnet'):
        return True
    physnet = vif.physnet
    mapping_res = CONF.sriov_physnet_resource_mappings
    try:
        resource = mapping_res[physnet]
    except KeyError:
        LOG.exception("No resource name for physnet %s", physnet)
        raise
    mapping_driver = CONF.sriov_resource_driver_mappings
    try:
        driver_name = mapping_driver[resource]
    except KeyError:
        LOG.exception("No driver for resource_name %s", resource)
        raise
    if driver_name in consts.USERSPACE_DRIVERS:
        LOG.info("_configure_l3 will not be called for vif %s "
                 "because of it's driver", vif)
        return False
    return True


def connect(vif, instance_info, ifname, netns=None,
            is_default_gateway=True, container_id=None):
    netns = utils.convert_netns(netns)
    driver = _get_binding_driver(vif)
    os_vif.plug(vif, instance_info)
    driver.connect(vif, ifname, netns, container_id)
    if _need_configure_l3(vif):
        vif_dict = utils.osvif_vif_to_dict(vif)
        _configure_l3(vif_dict, ifname, netns, is_default_gateway)


def disconnect(vif, instance_info, ifname, netns=None,
               container_id=None, **kwargs):
    driver = _get_binding_driver(vif)
    driver.disconnect(vif, ifname, netns, container_id)
    os_vif.unplug(vif, instance_info)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from zun.objects import base
from zun.objects import fields as zun_fields


@base.ZunObjectRegistry.register
class RequestGroup(base.ZunPersistentObject, base.ZunObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'use_same_provider': fields.BooleanField(default=True),
        'resources': fields.DictOfIntegersField(default={}),
        'required_traits': zun_fields.SetOfStringsField(default=set()),
        'forbidden_traits': zun_fields.SetOfStringsField(default=set()),
        # The aggregates field has a form of
        #     [[aggregate_UUID1],
        #      [aggregate_UUID2, aggregate_UUID3]]
        # meaning that the request should be fulfilled from an RP that is a
        # member of the aggregate aggregate_UUID1 and member of the aggregate
        # aggregate_UUID2 or aggregate_UUID3 .
        'aggregates': zun_fields.ListOfListsOfStringsField(default=[]),
        # The entity the request is coming from (e.g. the Neutron port uuid)
        # which may not always be a UUID.
        'requester_id': fields.StringField(nullable=True, default=None),
        # The resource provider UUIDs that together fulfill the request
        # NOTE(gibi): this can be more than one if this is the unnumbered
        # request group (i.e. use_same_provider=False)
        'provider_uuids': fields.ListOfUUIDField(default=[]),
        'in_tree': fields.UUIDField(nullable=True, default=None),
    }

    def __init__(self, context=None, **kwargs):
        super(RequestGroup, self).__init__(context=context, **kwargs)
        self.obj_set_defaults()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import yaml


def load(s):
    try:
        yml_dict = yaml.safe_load(s)
    except yaml.YAMLError as exc:
        msg = 'An error occurred during YAML parsing.'
        if hasattr(exc, 'problem_mark'):
            msg += ' Error position: (%s:%s)' % (exc.problem_mark.line + 1,
                                                 exc.problem_mark.column + 1)
        raise ValueError(msg)
    if not isinstance(yml_dict, dict) and not isinstance(yml_dict, list):
        raise ValueError('The source is not a YAML mapping or list.')
    if isinstance(yml_dict, dict) and len(yml_dict) < 1:
        raise ValueError('Could not find any element in your YAML mapping.')
    return yml_dict


def dump(s):
    return yaml.safe_dump(s)



# Copyright (c) 2014 Docker, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova import utils


def sys_uptime():
    """Returns the host uptime."""
    out, err = utils.execute('env', 'LANG=C', 'uptime')
    return out



# Copyright (c) 2017 Intel, Inc.
# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils

from zun.common import exception
from zun import objects
from zun.objects import fields
from zun.pci import stats
from zun.pci import whitelist

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class PciDevTracker(object):
    """Manage pci devices in a compute node.

    This class fetches pci passthrough information from compute node
    and tracks the usage of these devices.

    It's called by compute node resource tracker to allocate and free
    devices to/from containers, and to update the available pci passthrough
    devices information from compute node periodically.

    `pci_devs` attribute of this class is the in-memory "master copy" of all
    devices on each compute node, and all data changes that happen when
    claiming/allocating/freeing
    devices HAVE TO be made against container contained in `pci_devs` list,
    because they are periodically flushed to the DB when the save()
    method is called.

    It is unsafe to fetch PciDevice objects elsewhere in the code for update
    purposes as those changes will end up being overwritten when the `pci_devs`
    are saved.
    """

    def __init__(self, context, node_id=None):
        """Create a pci device tracker.

        If a node_id is passed in, it will fetch pci devices information
        from database, otherwise, it will create an empty devices list
        and the resource tracker will update the node_id information later.
        """

        super(PciDevTracker, self).__init__()
        self.stale = {}
        self.node_id = node_id
        self.dev_filter = whitelist.Whitelist(CONF.pci.passthrough_whitelist)
        self.stats = stats.PciDeviceStats(dev_filter=self.dev_filter)
        self._context = context
        if node_id:
            self.pci_devs = objects.PciDevice.list_by_compute_node(
                context, node_id)
        else:
            self.pci_devs = []
        self._build_device_tree(self.pci_devs)
        self._initial_instance_usage()

    def _initial_instance_usage(self):
        self.allocations = collections.defaultdict(list)
        self.claims = collections.defaultdict(list)
        for dev in self.pci_devs:
            uuid = dev.container_uuid
            if dev.status == fields.PciDeviceStatus.CLAIMED:
                self.claims[uuid].append(dev)
            elif dev.status == fields.PciDeviceStatus.ALLOCATED:
                self.allocations[uuid].append(dev)
            elif dev.status == fields.PciDeviceStatus.AVAILABLE:
                self.stats.add_device(dev)

    def save(self):
        for dev in self.pci_devs:
            if dev.obj_what_changed():
                dev.save()
                if dev.status == fields.PciDeviceStatus.DELETED:
                    self.pci_devs.remove(dev)

    @property
    def pci_stats(self):
        return self.stats

    def update_devices_from_compute_resources(self, devices_json):
        """Sync the pci device tracker with compute node information.

        To support pci device hot plug, we sync with the compute node
        periodically, fetching all devices information from compute node,
        update the tracker and sync the DB information.

        Devices should not be hot-plugged when assigned to a container,
        but possibly the compute node has no such guarantee. The best
        we can do is to give a warning if a device is changed
        or removed while assigned.

        :param devices_json: The JSON-ified string of device information
                             that is returned from the compute node.
        """

        devices = []
        for dev in jsonutils.loads(devices_json):
            if self.dev_filter.device_assignable(dev):
                devices.append(dev)
        self._set_hvdevs(devices)

    @staticmethod
    def _build_device_tree(all_devs):
        """Build a tree of devices that represents parent-child relationships.

        We need to have the relationships set up so that we can easily make
        all the necessary changes to parent/child devices without having to
        figure it out at each call site.

        This method just adds references to relevant containers already found
        in `pci_devs` to `child_devices` and `parent_device` fields of each
        one.

        Currently relationships are considered for SR-IOV PFs/VFs only.
        """

        # Ensures that devices are ordered in ASC so VFs will come
        # after their PFs.
        all_devs.sort(key=lambda x: x.address)

        parents = {}
        for dev in all_devs:
            if dev.status in (fields.PciDeviceStatus.REMOVED,
                              fields.PciDeviceStatus.DELETED):
                # NOTE(ndipanov): Removed devs are pruned from
                # self.pci_devs on save() so we need to make sure we
                # are not looking at removed ones as we may build up
                # the tree sooner than they are pruned.
                continue
            if dev.dev_type == fields.PciDeviceType.SRIOV_PF:
                dev.child_devices = []
                parents[dev.address] = dev
            elif dev.dev_type == fields.PciDeviceType.SRIOV_VF:
                dev.parent_device = parents.get(dev.parent_addr)
                if dev.parent_device:
                    parents[dev.parent_addr].child_devices.append(dev)

    def _set_hvdevs(self, devices):
        exist_addrs = set([dev.address for dev in self.pci_devs])
        new_addrs = set([dev['address'] for dev in devices])

        for existed in self.pci_devs:
            if existed.address in exist_addrs - new_addrs:
                try:
                    existed.remove()
                except exception.PciDeviceInvalidStatus as e:
                    LOG.warning(("Trying to remove device with %(status)s "
                                 "ownership %(instance_uuid)s because of "
                                 "%(pci_exception)s"),
                                {'status': existed.status,
                                 'container_uuid': existed.container_uuid,
                                 'pci_exception': e.format_message()})
                    # Note(yjiang5): remove the device by force so that
                    # db entry is cleaned in next sync.
                    existed.status = fields.PciDeviceStatus.REMOVED
                else:
                    # Note(yjiang5): no need to update stats if an assigned
                    # device is hot removed.
                    self.stats.remove_device(existed)
            else:
                new_value = next((dev for dev in devices if
                                 dev['address'] == existed.address))
                new_value['compute_node_id'] = self.node_id
                if existed.status in (fields.PciDeviceStatus.CLAIMED,
                                      fields.PciDeviceStatus.ALLOCATED):
                    # Pci properties may change while assigned because of
                    # hotplug or config changes. Although normally this should
                    # not happen.

                    # As the devices have been assigned to a container,
                    # we defer the change till the container is destroyed.
                    # We will not sync the new properties with database
                    # before that.

                    # TODO(yjiang5): Not sure if this is a right policy, but
                    # at least it avoids some confusion and, if needed,
                    # we can add more action like killing the container
                    # by force in future.
                    self.stale[new_value['address']] = new_value
                else:
                    existed.update_device(new_value)

        for dev in [dev for dev in devices if
                    dev['address'] in new_addrs - exist_addrs]:
            dev['compute_node_uuid'] = self.node_id
            dev_obj = objects.PciDevice.create(self._context, dev)
            self.pci_devs.append(dev_obj)
            self.stats.add_device(dev_obj)

        self._build_device_tree(self.pci_devs)

    def _claim_container(self, context, container_uuid, pci_requests):
        devs = self.stats.consume_requests(pci_requests.requests)
        if not devs:
            return None

        for dev in devs:
            dev.claim(container_uuid)
        return devs

    def _allocate_container(self, container, devs):
        for dev in devs:
            dev.allocate(container)

    def allocate_container(self, container):
        devs = self.claims.pop(container.uuid, [])
        self._allocate_container(container, devs)
        if devs:
            self.allocations[container.uuid] += devs

    def claim_container(self, context, container_uuid, pci_requests):
        devs = []
        if self.pci_devs and pci_requests and pci_requests.requests:
            devs = self._claim_container(context, container_uuid, pci_requests)
            if devs:
                self.claims[container_uuid] = devs
        return devs

    def free_device(self, dev, container):
        """Free device from pci resource tracker

        :param dev: cloned pci device object that needs to be free
        :param container: the container that this pci device
                          is allocated to
        """
        for pci_dev in self.pci_devs:
            # Find the matching pci device in the pci resource tracker.
            # Once found, free it.
            if (dev.id == pci_dev.id and
                    dev.container_uuid == container.uuid):
                self._remove_device_from_pci_mapping(
                    container.uuid, pci_dev, self.allocations)
                self._remove_device_from_pci_mapping(
                    container.uuid, pci_dev, self.claims)
                self._free_device(pci_dev)
                break

    def _remove_device_from_pci_mapping(self, container_uuid,
                                        pci_device, pci_mapping):
        """Remove a PCI device from allocations or claims.

        If there are no more PCI devices, pop the uuid.
        """
        pci_devices = pci_mapping.get(container_uuid, [])
        if pci_device in pci_devices:
            pci_devices.remove(pci_device)
            if len(pci_devices) == 0:
                pci_mapping.pop(container_uuid, None)

    def _free_device(self, dev, container=None):
        freed_devs = dev.free(container)
        stale = self.stale.pop(dev.address, None)
        if stale:
            dev.update_device(stale)
        for dev in freed_devs:
            self.stats.add_device(dev)

    def _free_container(self, container):
        for dev in self.pci_devs:
            if dev.status in (fields.PciDeviceStatus.CLAIMED,
                              fields.PciDeviceStatus.ALLOCATED):
                if dev.container_uuid == container.uuid:
                    self._free_device(dev)

    def free_container(self, context, container):
        if self.allocations.pop(container.uuid, None):
            self._free_container(container)
        elif self.claims.pop(container.uuid, None):
            self._free_container(container)

    def update_pci_for_container(self, context, container, sign):
        """Update PCI usage information if devices are de/allocated."""
        if not self.pci_devs:
            return

        if sign == -1:
            self.free_container(context, container)
        if sign == 1:
            self.allocate_container(container)

    def clean_usage(self, containers, orphans):
        """Remove all usages for containers not passed in the parameter.

        The caller should hold the COMPUTE_RESOURCE_SEMAPHORE lock
        """
        existed = set(cnt['uuid'] for cnt in containers)
        existed |= set(cnt['uuid'] for cnt in orphans)

        # need to copy keys, because the dict is modified in the loop body
        for uuid in list(self.claims):
            if uuid not in existed:
                devs = self.claims.pop(uuid, [])
                for dev in devs:
                    self._free_device(dev)
        # need to copy keys, because the dict is modified in the loop body
        for uuid in list(self.allocations):
            if uuid not in existed:
                devs = self.allocations.pop(uuid, [])
                for dev in devs:
                    self._free_device(dev)


def get_container_pci_devs(cnt, request_id=None):
    """Get the devices allocated to one or all requests for a container.

    - For generic PCI request, the request id is None.
    - For sr-iov networking, the request id is a valid uuid
    - There are a couple of cases where all the PCI devices allocated to a
      container need to be returned.
    """
    pci_devices = cnt.pci_devices
    if pci_devices is None:
        return []
    return [device for device in pci_devices
            if device.request_id == request_id or request_id == 'all']



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zun.common.docker_image import digest as digest_
from zun.common.docker_image import regexp
from zun.common import exception


ImageRegexps = regexp.ImageRegexps

NAME_TOTAL_LENGTH_MAX = 255


class Reference(dict):
    def __init__(self, name=None, tag=None, digest=None):
        super(Reference, self).__init__()
        self['name'] = name
        self['tag'] = tag
        self['digest'] = digest

    def split_hostname(self):
        name = self['name']
        matched = ImageRegexps.ANCHORED_NAME_REGEXP.match(name)

        if not matched:
            return '', name
        matches = matched.groups()
        if len(matches) != 2:
            return '', name
        hostname = matches[0]
        if hostname is None:
            return '', name
        if ('.' not in hostname and ':' not in hostname and
                hostname != 'localhost'):
            return '', name

        return matches[0], matches[1]

    def string(self):
        return '{}:{}@{}'.format(self['name'], self['tag'], self['digest'])

    def best_reference(self):
        if not self['name']:
            if self['digest']:
                return DigestReference(self['digest'])
            return None

        if not self['tag']:
            if self['digest']:
                return CanonicalReference(self['name'], self['digest'])
            return NamedReference(self['name'])

        if not self['digest']:
            return TaggedReference(self['name'], self['tag'])

        return self

    @classmethod
    def parse(cls, s):
        matched = ImageRegexps.REFERENCE_REGEXP.match(s)
        if not matched and not s:
            raise exception.NameEmpty()
        if not matched:
            raise exception.ReferenceInvalidFormat()

        matches = matched.groups()
        if len(matches[0]) > NAME_TOTAL_LENGTH_MAX:
            raise exception.NameTooLong(length_max=NAME_TOTAL_LENGTH_MAX)

        ref = cls(name=matches[0], tag=matches[1])
        if matches[2]:
            digest_.validate_digest(matches[2])
            ref['digest'] = matches[2]

        r = ref.best_reference()
        if not r:
            raise exception.NameEmpty()

        return r


class NamedReference(Reference):
    def __init__(self, name, **kwargs):
        super(NamedReference, self).__init__(name=name, **kwargs)

    def string(self):
        return '{}'.format(self['name'])


class DigestReference(Reference):
    def __init__(self, digest, **kwargs):
        super(DigestReference, self).__init__(digest=digest, **kwargs)

    def string(self):
        return self['digest']


class CanonicalReference(NamedReference):
    def __init__(self, name, digest, **kwargs):
        super(CanonicalReference, self).__init__(name=name, digest=digest,
                                                 **kwargs)

    def string(self):
        return '{}@{}'.format(self['name'], self['digest'])


class TaggedReference(NamedReference):
    def __init__(self, name, tag, **kwargs):
        super(TaggedReference, self).__init__(name=name, tag=tag, **kwargs)

    def string(self):
        return '{}:{}'.format(self['name'], self['tag'])



# Copyright 2016 Intel.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from oslo_log import log as logging
import stevedore

from zun.common.i18n import _
import zun.conf

CONF = zun.conf.CONF
LOG = logging.getLogger(__name__)


def load_image_driver(image_driver=None):
    """Load an image driver module.

    Load the container image driver module specified by the image_driver
    configuration option or, if supplied, the driver name supplied as an
    argument.
    :param image_driver: container image driver name to override config opt
    :returns: a ContainerImageDriver instance
    """
    if not image_driver:
        image_driver = CONF.default_image_driver
    if not image_driver:
        LOG.error("Container image driver option required, "
                  "but not specified")
        sys.exit(1)

    LOG.info("Loading container image driver '%s'", image_driver)
    try:
        driver = stevedore.driver.DriverManager(
            "zun.image.driver",
            image_driver,
            invoke_on_load=True).driver

        if not isinstance(driver, ContainerImageDriver):
            raise Exception(_('Expected driver of type: %s') %
                            str(ContainerImageDriver))

        return driver
    except Exception:
        LOG.exception("Unable to load the container image driver")
        sys.exit(1)


class ContainerImageDriver(object):
    """Base class for container image driver."""

    def pull_image(self, context, repo, tag, image_pull_policy, registry):
        """Pull an image."""
        raise NotImplementedError()

    def search_image(self, context, repo, tag, exact_match):
        """Search an image."""
        raise NotImplementedError()

    def create_image(self, context, image_name):
        """Create an image."""
        raise NotImplementedError()

    def update_image(self, context, img_id, container_fmt=None,
                     disk_fmt=None, tag=None):
        """Update an image."""
        raise NotImplementedError()

    def upload_image_data(self, context, img_id, data):
        """Upload an image."""
        raise NotImplementedError()

    def delete_image(self, context, img_id):
        """Delete an image."""
        raise NotImplementedError()

    def delete_committed_image(self, context, img_id, image_driver):
        """Delete a committed image."""
        raise NotImplementedError()

    def delete_image_tar(self, context, image):
        """Delete an image."""
        raise NotImplementedError()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating resource providers via the DB API"""

from oslo_utils import uuidutils

from zun.common import exception
import zun.conf
from zun.db import api as dbapi
from zun.tests.unit.db import base
from zun.tests.unit.db import utils

CONF = zun.conf.CONF


class DbResourceProviderTestCase(base.DbTestCase):

    def setUp(self):
        super(DbResourceProviderTestCase, self).setUp()

    def test_create_resource_provider(self):
        utils.create_test_resource_provider(context=self.context)

    def test_create_resource_provider_already_exists(self):
        utils.create_test_resource_provider(
            context=self.context, uuid='123')
        with self.assertRaisesRegex(exception.ResourceProviderAlreadyExists,
                                    'A resource provider with UUID 123.*'):
            utils.create_test_resource_provider(
                context=self.context, uuid='123')

    def test_get_resource_provider_by_uuid(self):
        provider = utils.create_test_resource_provider(context=self.context)
        res = dbapi.get_resource_provider(
            self.context, provider.uuid)
        self.assertEqual(provider.id, res.id)
        self.assertEqual(provider.uuid, res.uuid)

    def test_get_resource_provider_by_name(self):
        provider = utils.create_test_resource_provider(context=self.context)
        res = dbapi.get_resource_provider(
            self.context, provider.name)
        self.assertEqual(provider.id, res.id)
        self.assertEqual(provider.uuid, res.uuid)

    def test_get_resource_provider_that_does_not_exist(self):
        self.assertRaises(exception.ResourceProviderNotFound,
                          dbapi.get_resource_provider,
                          self.context,
                          uuidutils.generate_uuid())

    def test_list_resource_providers(self):
        uuids = []
        for i in range(1, 6):
            provider = utils.create_test_resource_provider(
                uuid=uuidutils.generate_uuid(),
                context=self.context,
                name='provider' + str(i))
            uuids.append(str(provider['uuid']))
        res = dbapi.list_resource_providers(self.context)
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), sorted(res_uuids))

    def test_list_resource_providers_sorted(self):
        uuids = []
        for i in range(5):
            provider = utils.create_test_resource_provider(
                uuid=uuidutils.generate_uuid(),
                context=self.context,
                name='provider' + str(i))
            uuids.append(str(provider.uuid))
        res = dbapi.list_resource_providers(self.context, sort_key='uuid')
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), res_uuids)

        self.assertRaises(exception.InvalidParameterValue,
                          dbapi.list_resource_providers,
                          self.context,
                          sort_key='foo')

    def test_list_resource_providers_with_filters(self):
        provider1 = utils.create_test_resource_provider(
            name='provider-one',
            uuid=uuidutils.generate_uuid(),
            context=self.context)
        provider2 = utils.create_test_resource_provider(
            name='provider-two',
            uuid=uuidutils.generate_uuid(),
            context=self.context)

        res = dbapi.list_resource_providers(
            self.context, filters={'name': 'provider-one'})
        self.assertEqual([provider1.id], [r.id for r in res])

        res = dbapi.list_resource_providers(
            self.context, filters={'name': 'provider-two'})
        self.assertEqual([provider2.id], [r.id for r in res])

        res = dbapi.list_resource_providers(
            self.context, filters={'name': 'bad-provider'})
        self.assertEqual([], [r.id for r in res])

        res = dbapi.list_resource_providers(
            self.context,
            filters={'name': provider1.name})
        self.assertEqual([provider1.id], [r.id for r in res])

    def test_destroy_resource_provider(self):
        provider = utils.create_test_resource_provider(context=self.context)
        dbapi.destroy_resource_provider(self.context, provider.id)
        self.assertRaises(exception.ResourceProviderNotFound,
                          dbapi.get_resource_provider,
                          self.context, provider.uuid)

    def test_destroy_resource_provider_by_uuid(self):
        provider = utils.create_test_resource_provider(context=self.context)
        dbapi.destroy_resource_provider(self.context, provider.uuid)
        self.assertRaises(exception.ResourceProviderNotFound,
                          dbapi.get_resource_provider,
                          self.context, provider.uuid)

    def test_destroy_resource_provider_that_does_not_exist(self):
        self.assertRaises(exception.ResourceProviderNotFound,
                          dbapi.destroy_resource_provider, self.context,
                          uuidutils.generate_uuid())

    def test_update_resource_provider(self):
        provider = utils.create_test_resource_provider(context=self.context)
        old_name = provider.name
        new_name = 'new-name'
        self.assertNotEqual(old_name, new_name)

        res = dbapi.update_resource_provider(
            self.context, provider.id, {'name': new_name})
        self.assertEqual(new_name, res.name)

    def test_update_resource_provider_not_found(self):
        provider_uuid = uuidutils.generate_uuid()
        new_name = 'new-name'
        self.assertRaises(exception.ResourceProviderNotFound,
                          dbapi.update_resource_provider, self.context,
                          provider_uuid, {'name': new_name})

    def test_update_resource_provider_uuid(self):
        provider = utils.create_test_resource_provider(context=self.context)
        self.assertRaises(exception.InvalidParameterValue,
                          dbapi.update_resource_provider, self.context,
                          provider.id, {'uuid': ''})



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add-missing-index-and-foreign-key

Revision ID: 74c97dca93d0
Revises: 47d79ffdc582
Create Date: 2020-03-27 02:51:30.959226

"""

# revision identifiers, used by Alembic.
revision = '74c97dca93d0'
down_revision = '47d79ffdc582'
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('container', schema=None) as batch_op:
        batch_op.create_foreign_key(
            None, 'container', ['capsule_id'], ['id'], ondelete='CASCADE')
    # ### end Alembic commands ###



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import functools
import shutil

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import excutils
from oslo_utils import fileutils
from stevedore import driver as stevedore_driver

from zun.common import exception
from zun.common.i18n import _
from zun.common import mount
from zun.common import utils
import zun.conf
from zun.volume import cinder_api
from zun.volume import cinder_workflow

LOG = logging.getLogger(__name__)

CONF = zun.conf.CONF


def driver(driver_name, *args, **kwargs):
    LOG.info("Loading volume driver '%s'", driver_name)
    volume_driver = stevedore_driver.DriverManager(
        "zun.volume.driver",
        driver_name,
        invoke_on_load=True,
        invoke_args=args,
        invoke_kwds=kwargs).driver
    if not isinstance(volume_driver, VolumeDriver):
        raise exception.ZunException(_("Invalid volume driver type"))
    return volume_driver


def validate_volume_provider(supported_providers):
    """Wraps a method to validate volume provider."""

    def decorator(function):
        @functools.wraps(function)
        def decorated_function(self, context, volume, **kwargs):
            provider = volume.volume_provider
            if provider not in supported_providers:
                msg = _("The volume provider '%s' is not supported") % provider
                raise exception.ZunException(msg)

            return function(self, context, volume, **kwargs)

        return decorated_function
    return decorator


class VolumeDriver(object, metaclass=abc.ABCMeta):
    """The base class that all Volume classes should inherit from."""

    def attach(self, *args, **kwargs):
        raise NotImplementedError()

    def detach(self, *args, **kwargs):
        raise NotImplementedError()

    def delete(self, *args, **kwargs):
        raise NotImplementedError()

    def bind_mount(self, *args, **kwargs):
        raise NotImplementedError()

    def is_volume_available(self, context, volmap):
        raise NotImplementedError()

    def is_volume_deleted(self, context, volmap):
        raise NotImplementedError()


class Local(VolumeDriver):

    supported_providers = ['local']

    @validate_volume_provider(supported_providers)
    def attach(self, context, volmap):
        mountpoint = mount.get_mountpoint(volmap.volume.uuid)
        fileutils.ensure_tree(mountpoint)
        filename = '/'.join([mountpoint, volmap.volume.uuid])
        with open(filename, 'wb') as fd:
            content = utils.decode_file_data(volmap.contents)
            fd.write(content)

    def _remove_local_file(self, volmap):
        mountpoint = mount.get_mountpoint(volmap.volume.uuid)
        shutil.rmtree(mountpoint)

    @validate_volume_provider(supported_providers)
    def detach(self, context, volmap):
        self._remove_local_file(volmap)

    @validate_volume_provider(supported_providers)
    def delete(self, context, volmap):
        self._remove_local_file(volmap)

    @validate_volume_provider(supported_providers)
    def bind_mount(self, context, volmap):
        mountpoint = mount.get_mountpoint(volmap.volume.uuid)
        filename = '/'.join([mountpoint, volmap.volume.uuid])
        return filename, volmap.container_path

    def is_volume_available(self, context, volmap):
        return True, False

    def is_volume_deleted(self, context, volmap):
        return True, False


class Cinder(VolumeDriver):

    supported_providers = [
        'cinder'
    ]

    @validate_volume_provider(supported_providers)
    def attach(self, context, volmap):
        cinder = cinder_workflow.CinderWorkflow(context)
        if volmap.connection_info:
            # this is a re-attach of the volume
            connection_info = jsonutils.loads(volmap.connection_info)
            device_info = cinder._connect_volume(connection_info)
            connection_info['data']['device_path'] = device_info['path']
            try:
                volmap.connection_info = jsonutils.dumps(connection_info)
            except TypeError:
                pass
            volmap.save(context)
            devpath = connection_info['data']['device_path']
        else:
            # this is the first time to attach the volume
            devpath = cinder.attach_volume(volmap)

        try:
            self._mount_device(volmap, devpath)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("Failed to mount device")
                try:
                    cinder.detach_volume(context, volmap)
                except Exception:
                    LOG.exception("Failed to detach volume")

    def _mount_device(self, volmap, devpath):
        mountpoint = mount.get_mountpoint(volmap.volume.uuid)
        fileutils.ensure_tree(mountpoint)
        mount.do_mount(devpath, mountpoint, CONF.volume.fstype)

    @validate_volume_provider(supported_providers)
    def detach(self, context, volmap):
        self._unmount_device(volmap)
        cinder = cinder_workflow.CinderWorkflow(context)
        cinder.detach_volume(context, volmap)

    @validate_volume_provider(supported_providers)
    def delete(self, context, volmap):
        cinder = cinder_workflow.CinderWorkflow(context)
        cinder.delete_volume(volmap)

    def _unmount_device(self, volmap):
        if hasattr(volmap, 'connection_info'):
            mountpoint = mount.get_mountpoint(volmap.volume.uuid)
            mount.do_unmount(mountpoint)
            shutil.rmtree(mountpoint)

    @validate_volume_provider(supported_providers)
    def bind_mount(self, context, volmap):
        mountpoint = mount.get_mountpoint(volmap.volume.uuid)
        return mountpoint, volmap.container_path

    @validate_volume_provider(supported_providers)
    def get_volume_status(self, context, volmap):
        ca = cinder_api.CinderAPI(context)
        return ca.get(volmap.cinder_volume_id).status

    @validate_volume_provider(supported_providers)
    def check_multiattach(self, context, volmap):
        ca = cinder_api.CinderAPI(context)
        return ca.get(volmap.cinder_volume_id).multiattach

    @validate_volume_provider(supported_providers)
    def is_volume_available(self, context, volmap):
        status = self.get_volume_status(context, volmap)
        if status == 'available':
            is_available = True
            is_error = False
        elif status == 'in-use':
            multiattach = self.check_multiattach(context, volmap)
            is_available = multiattach
            is_error = False
        elif status == 'error':
            is_available = False
            is_error = True
        else:
            is_available = False
            is_error = False

        return is_available, is_error

    @validate_volume_provider(supported_providers)
    def is_volume_deleted(self, context, volmap):
        try:
            volume = cinder_api.CinderAPI(context).search_volume(
                volmap.cinder_volume_id)
            is_deleted = False
            # Cinder volume error states: 'error', 'error_deleting',
            # 'error_backing-up', 'error_restoring', 'error_extending',
            # all of which start with 'error'
            is_error = True if 'error' in volume.status else False
        except exception.VolumeNotFound:
            is_deleted = True
            is_error = False

        return is_deleted, is_error



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_versionedobjects import fields

from zun.common import exception
from zun.db import api as dbapi
from zun.objects import base
from zun.objects import container
from zun.objects import volume as volume_obj


LOG = logging.getLogger(__name__)


_VOLUME_MAPPING_OPTIONAL_JOINED_FIELDS = [
    'container',
    'volume',
]
VOLUME_ATTRS = [
    'volume_provider',
    'cinder_volume_id',
    'connection_info',
    'auto_remove',
    'host',
    'contents',
]
VOLUME_MAPPING_OPTIONAL_ATTRS = \
    _VOLUME_MAPPING_OPTIONAL_JOINED_FIELDS + VOLUME_ATTRS


def _expected_cols(expected_attrs):
    return [attr for attr in expected_attrs
            if attr in _VOLUME_MAPPING_OPTIONAL_JOINED_FIELDS]


@base.ZunObjectRegistry.register
class VolumeMapping(base.ZunPersistentObject, base.ZunObject):
    # Version 1.0: Initial version
    # Version 1.1: Add field "auto_remove"
    # Version 1.2: Add field "host"
    # Version 1.3: Add field "contents"
    # Version 1.4: Rename field "volume_id" to "cinder_volume_id"
    # Version 1.5: Add method "count"
    VERSION = '1.5'

    fields = {
        'id': fields.IntegerField(),
        'uuid': fields.UUIDField(nullable=False),
        'project_id': fields.StringField(nullable=True),
        'user_id': fields.StringField(nullable=True),
        'cinder_volume_id': fields.UUIDField(nullable=True),
        'volume_provider': fields.StringField(nullable=False),
        'container_path': fields.StringField(nullable=True),
        'container_uuid': fields.UUIDField(nullable=True),
        'container': fields.ObjectField('ContainerBase', nullable=True),
        'connection_info': fields.SensitiveStringField(nullable=True),
        'auto_remove': fields.BooleanField(nullable=True),
        'host': fields.StringField(nullable=True),
        'contents': fields.SensitiveStringField(nullable=True),
        'volume_id': fields.IntegerField(nullable=False),
        'volume': fields.ObjectField('Volume', nullable=True),
    }

    @staticmethod
    def _from_db_object(volume, db_volume):
        """Converts a database entity to a formal object."""
        for field in volume.fields:
            if field in VOLUME_MAPPING_OPTIONAL_ATTRS:
                continue
            setattr(volume, field, db_volume[field])

        volume.obj_reset_changes()
        return volume

    @staticmethod
    def _from_db_object_list(db_objects, cls, context):
        """Converts a list of database entities to a list of formal objects."""
        return [VolumeMapping._from_db_object(cls(context), obj)
                for obj in db_objects]

    @base.remotable_classmethod
    def get_by_uuid(cls, context, uuid):
        """Find a volume mapping based on uuid.

        :param uuid: the uuid of a volume mapping.
        :param context: Security context
        :returns: a :class:`VolumeMapping` object.
        """
        db_volume = dbapi.get_volume_mapping_by_uuid(context, uuid)
        volume = VolumeMapping._from_db_object(cls(context), db_volume)
        return volume

    @base.remotable_classmethod
    def list(cls, context, limit=None, marker=None,
             sort_key=None, sort_dir=None, filters=None):
        """Return a list of VolumeMapping objects.

        :param context: Security context.
        :param limit: maximum number of resources to return in a single result.
        :param marker: pagination marker for large data sets.
        :param sort_key: column to sort results by.
        :param sort_dir: direction to sort. "asc" or "desc".
        :param filters: filters when list volume mappings.
        :returns: a list of :class:`VolumeMapping` object.

        """
        db_volumes = dbapi.list_volume_mappings(
            context, limit=limit, marker=marker, sort_key=sort_key,
            sort_dir=sort_dir, filters=filters)
        return VolumeMapping._from_db_object_list(db_volumes, cls, context)

    @base.remotable_classmethod
    def list_by_container(cls, context, container_uuid):
        filters = {'container_uuid': container_uuid}
        db_volumes = dbapi.list_volume_mappings(context, filters=filters)
        return VolumeMapping._from_db_object_list(db_volumes, cls, context)

    @base.remotable_classmethod
    def list_by_cinder_volume(cls, context, cinder_volume_id):
        filters = {'cinder_volume_id': cinder_volume_id}
        db_volumes = dbapi.list_volume_mappings(context, filters=filters)
        return VolumeMapping._from_db_object_list(db_volumes, cls, context)

    @base.remotable_classmethod
    def count(cls, context, **filters):
        return dbapi.count_volume_mappings(context, **filters)

    @base.remotable
    def create(self, context):
        """Create a VolumeMapping record in the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object.

        """
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason='already created')
        values = self.obj_get_changes()
        if 'container' in values:
            raise exception.ObjectActionError(action='create',
                                              reason='container assigned')
        if 'volume' in values:
            raise exception.ObjectActionError(action='create',
                                              reason='volume assigned')

        self._create_volume(context, values)
        db_volume = dbapi.create_volume_mapping(context, values)
        self._from_db_object(self, db_volume)

    def _create_volume(self, context, values):
        volume_values = {}
        for attrname in list(values.keys()):
            if attrname in VOLUME_ATTRS:
                volume_values[attrname] = values.pop(attrname)
        volume_values['user_id'] = values['user_id']
        volume_values['project_id'] = values['project_id']
        if 'volume_id' not in values:
            volume = volume_obj.Volume(context, **volume_values)
            volume.create(context)
            values['volume_id'] = volume.id

    @base.remotable
    def destroy(self, context=None):
        """Delete the VolumeMapping from the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object.
        """
        context = context or self._context
        if not self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='destroy',
                                              reason='already destroyed')
        dbapi.destroy_volume_mapping(context, self.uuid)
        self._destroy_volume(context)
        delattr(self, 'id')
        self.obj_reset_changes()

    def _destroy_volume(self, context):
        if VolumeMapping.count(context,
                               volume_id=self.volume_id) == 0:
            dbapi.destroy_volume(context, self.volume_id)

    @base.remotable
    def save(self, context=None):
        """Save updates to this VolumeMapping.

        Updates will be made column by column based on the result
        of self.what_changed().

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object.
        """
        updates = self.obj_get_changes()
        if 'container' in updates:
            raise exception.ObjectActionError(action='save',
                                              reason='container changed')
        if 'volume' in updates:
            raise exception.ObjectActionError(action='save',
                                              reason='volume changed')
        updates.pop('id', None)
        self._update_volume(context, updates)
        dbapi.update_volume_mapping(context, self.uuid, updates)

        self.obj_reset_changes()

    def _update_volume(self, context, values):
        volume = self.volume
        for attrname in list(values.keys()):
            if attrname in VOLUME_ATTRS:
                setattr(volume, attrname, values.pop(attrname))
        volume.save(context)

    @base.remotable
    def refresh(self, context=None):
        """Loads updates for this VolumeMapping.

        Loads a volume mapping with the same uuid from the database and
        checks for updated attributes. Updates are applied from
        the loaded volume mapping column by column, if there are any updates.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object.
        """
        current = self.__class__.get_by_uuid(self._context, uuid=self.uuid)
        for field in self.fields:
            if not self.obj_attr_is_set(field):
                continue
            if field == 'volume':
                self.volume.refresh()
            elif field == 'container':
                self.container.refresh()
            elif getattr(self, field) != getattr(current, field):
                setattr(self, field, getattr(current, field))
        self.obj_reset_changes()

    def obj_load_attr(self, attrname):
        if attrname not in VOLUME_MAPPING_OPTIONAL_ATTRS:
            raise exception.ObjectActionError(
                action='obj_load_attr',
                reason='attribute %s not lazy-loadable' % attrname)
        if not self._context:
            raise exception.OrphanedObjectError(method='obj_load_attr',
                                                objtype=self.obj_name())

        LOG.debug("Lazy-loading '%(attr)s' on %(name)s",
                  {'attr': attrname,
                   'name': self.obj_name(),
                   })

        if attrname in VOLUME_ATTRS:
            value = getattr(self.volume, attrname)
            setattr(self, attrname, value)
            self.obj_reset_changes(fields=[attrname])
        if attrname == 'container':
            self.container = container.ContainerBase.get_container_any_type(
                self._context, self.container_uuid)
            self.obj_reset_changes(fields=['container'])
        if attrname == 'volume':
            self.volume = volume_obj.Volume.get_by_id(self._context,
                                                      self.volume_id)
            self.obj_reset_changes(fields=['volume'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add compute node table

Revision ID: eeac0d191f5a
Revises: 8192905fd835
Create Date: 2017-02-28 21:32:58.122924

"""

# revision identifiers, used by Alembic.
revision = 'eeac0d191f5a'
down_revision = '8192905fd835'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from zun.db.sqlalchemy import models


def upgrade():
    op.create_table(
        'compute_node',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('uuid', sa.String(length=36), nullable=False),
        sa.Column('hostname', sa.String(length=255), nullable=False),
        sa.Column('numa_topology', models.JSONEncodedDict(), nullable=True),
        sa.PrimaryKeyConstraint('uuid'),
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections

from zun.api.controllers import link as link_module
from zun.tests import base as test_base


class TestLink(test_base.BaseTestCase):

    def test_make_link(self):
        link = link_module.make_link(
            'self', 'http://localhost:8080', 'v1', '',
            bookmark=True)

        ordered_link = collections.OrderedDict(sorted(link.items()))
        expected_value = collections.OrderedDict([
            ('href', 'http://localhost:8080/v1/'),
            ('rel', 'self')
        ])
        self.assertEqual(str(expected_value),
                         str(ordered_link))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinderclient import exceptions as cinder_exception
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import uuidutils

from zun.common import clients
from zun.common import exception
from zun.common.i18n import _
import zun.conf


LOG = logging.getLogger(__name__)

CONF = zun.conf.CONF


class CinderAPI(object):

    def __init__(self, context):
        self.context = context
        self.cinder = clients.OpenStackClients(self.context).cinder()

    def __getattr__(self, key):
        return getattr(self.cinder, key)

    def get(self, volume_id):
        return self.cinder.volumes.get(volume_id)

    def search_volume(self, volume):
        if uuidutils.is_uuid_like(volume):
            try:
                volume = self.get(volume)
            except cinder_exception.NotFound:
                raise exception.VolumeNotFound(volume=volume)
        else:
            try:
                volume = self.cinder.volumes.find(name=volume)
            except cinder_exception.NotFound:
                raise exception.VolumeNotFound(volume=volume)
            except cinder_exception.NoUniqueMatch:
                raise exception.Conflict(_(
                    'Multiple cinder volumes exist with same name. '
                    'Please use the uuid instead.'))

        return volume

    def ensure_volume_usable(self, volume):
        # Make sure the container has access to the volume.
        if hasattr(volume, 'os-vol-tenant-attr:tenant_id'):
            project_id = self.context.project_id
            if getattr(volume, 'os-vol-tenant-attr:tenant_id') != project_id:
                raise exception.VolumeNotUsable(volume=volume.id)

        if volume.attachments and not volume.multiattach:
            raise exception.VolumeInUse(volume=volume.id)

    def reserve_volume(self, volume_id):
        return self.cinder.volumes.reserve(volume_id)

    def unreserve_volume(self, volume_id):
        return self.cinder.volumes.unreserve(volume_id)

    def initialize_connection(self, volume_id, connector):
        try:
            connection_info = self.cinder.volumes.initialize_connection(
                volume_id, connector)
            return connection_info
        except cinder_exception.ClientException as ex:
            with excutils.save_and_reraise_exception():
                LOG.error('Initialize connection failed for volume '
                          '%(vol)s on host %(host)s. Error: %(msg)s '
                          'Code: %(code)s. Attempting to terminate '
                          'connection.',
                          {'vol': volume_id,
                           'host': connector.get('host'),
                           'msg': str(ex),
                           'code': ex.code})
                try:
                    self.terminate_connection(volume_id, connector)
                except Exception as exc:
                    LOG.error('Connection between volume %(vol)s and host '
                              '%(host)s might have succeeded, but attempt '
                              'to terminate connection has failed. '
                              'Validate the connection and determine if '
                              'manual cleanup is needed. Error: %(msg)s '
                              'Code: %(code)s.',
                              {'vol': volume_id,
                               'host': connector.get('host'),
                               'msg': str(exc),
                               'code': (exc.code
                                        if hasattr(exc, 'code') else None)})

    def terminate_connection(self, volume_id, connector):
        return self.cinder.volumes.terminate_connection(volume_id, connector)

    def attach(self, volume_id, mountpoint, hostname, container_uuid=None):
        return self.cinder.volumes.attach(volume=volume_id,
                                          instance_uuid=container_uuid,
                                          mountpoint=mountpoint,
                                          host_name=hostname)

    def detach(self, volume_map):
        volume_id = volume_map.cinder_volume_id
        attachment_id = None
        volume = self.get(volume_id)
        attachments = volume.attachments or {}
        for am in attachments:
            if am['server_id'] == volume_map.container_uuid:
                attachment_id = am['attachment_id']
                break

        if attachment_id is None and volume.multiattach:
            LOG.warning("attachment_id couldn't be retrieved for "
                        "volume %(volume_id)s. The volume has the "
                        "'multiattach' flag enabled, without the "
                        "attachment_id Cinder most probably "
                        "cannot perform the detach.",
                        {'volume_id': volume_id})

        return self.cinder.volumes.detach(volume_id, attachment_id)

    def begin_detaching(self, volume_id):
        self.cinder.volumes.begin_detaching(volume_id)

    def roll_detaching(self, volume_id):
        self.cinder.volumes.roll_detaching(volume_id)

    def create_volume(self, size):
        try:
            volume = self.cinder.volumes.create(size)
        except cinder_exception.ClientException as ex:
            LOG.error('Volume creation failed: %(ex)s', {'ex': ex})
            raise exception.VolumeCreateFailed(creation_failed=ex)

        return volume

    def delete_volume(self, volume_id):
        try:
            self.cinder.volumes.delete(volume_id)
        except cinder_exception.ClientException as ex:
            LOG.error('Volume deletion failed: %(ex)s',
                      {'ex': ex})
            raise exception.VolumeDeleteFailed(deletion_failed=ex)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating Quota via the DB API"""

from zun.common import context
from zun.common import exception
import zun.conf
from zun.db import api as dbapi
from zun.tests.unit.db import base
from zun.tests.unit.db import utils

CONF = zun.conf.CONF


class DBQuotaTestCase(base.DbTestCase):

    def setUp(self):
        super(DBQuotaTestCase, self).setUp()
        self.ctx = context.get_admin_context()
        self.project_id = 'fake_project_id'
        self.resource = 'containers'
        self.limit = 100

    def test_create_quota(self):
        quota = utils.create_test_quota(context=self.ctx,
                                        project_id=self.project_id,
                                        resource=self.resource,
                                        limit=self.limit)
        self.assertEqual(quota.project_id, self.project_id)
        self.assertEqual(quota.resource, self.resource)
        self.assertEqual(quota.hard_limit, self.limit)

    def test_get_quota(self):
        quota = utils.create_test_quota(context=self.ctx,
                                        project_id=self.project_id,
                                        resource=self.resource,
                                        limit=self.limit)
        res = dbapi.quota_get(context=self.ctx,
                              project_id=quota.project_id,
                              resource=quota.resource)
        self.assertEqual(quota.project_id, res.project_id)
        self.assertEqual(quota.resource, res.resource)
        self.assertEqual(quota.hard_limit, res.hard_limit)

    def test_get_all_project_quota(self):
        quota_1 = utils.create_test_quota(context=self.ctx,
                                          project_id=self.project_id,
                                          resource='resource_1',
                                          limit=10)
        quota_2 = utils.create_test_quota(context=self.ctx,
                                          project_id=self.project_id,
                                          resource='resource_2',
                                          limit=20)
        quotas = dbapi.quota_get_all_by_project(self.ctx, self.project_id)
        self.assertEqual(quotas['project_id'], self.project_id)
        self.assertEqual(quotas[quota_1.resource], quota_1.hard_limit)
        self.assertEqual(quotas[quota_2.resource], quota_2.hard_limit)

    def test_destroy_quota(self):
        quota = utils.create_test_quota(context=self.ctx,
                                        project_id=self.project_id,
                                        resource=self.resource,
                                        limit=self.limit)
        dbapi.quota_destroy(self.ctx, quota.project_id, quota.resource)
        self.assertRaises(exception.ProjectQuotaNotFound, dbapi.quota_get,
                          self.ctx, quota.project_id, quota.resource)

    def test_update_quota(self):
        quota = utils.create_test_quota(context=self.ctx,
                                        project_id=self.project_id,
                                        resource=self.resource,
                                        limit=self.limit)
        dbapi.quota_update(self.ctx, quota.project_id,
                           quota.resource, 200)
        updated_quota = dbapi.quota_get(self.ctx, quota.project_id,
                                        quota.resource)
        self.assertEqual(updated_quota.hard_limit, 200)



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg


api_service_opts = [
    cfg.PortOpt('port',
                default=9517,
                help='The port for the zun API server.'),
    cfg.IPOpt('host_ip',
              default='$my_ip',
              help="The listen IP for the zun API server. "
                   "The default is ``$my_ip``, "
                   "the IP address of this host."),
    cfg.BoolOpt('enable_ssl_api',
                default=False,
                help="Enable the integrated stand-alone API to service "
                     "requests via HTTPS instead of HTTP. If there is a "
                     "front-end service performing HTTPS offloading from "
                     "the service, this option should be False; note, you "
                     "will want to change public API endpoint to represent "
                     "SSL termination URL with 'public_endpoint' option."),
    cfg.IntOpt('workers',
               help="Number of workers for zun-api service. "
                    "The default will be the number of CPUs available."),
    cfg.IntOpt('max_limit',
               default=1000,
               help='The maximum number of items returned in a single '
                    'response from a collection resource.'),
    cfg.StrOpt('api_paste_config',
               default="api-paste.ini",
               help="Configuration file for WSGI definition of API."),
    cfg.BoolOpt('enable_image_validation',
                default=False,
                help="Enable image validation.")
]


api_group = cfg.OptGroup(name='api',
                         title='Options for the zun-api service')


ALL_OPTS = (api_service_opts)


def register_opts(conf):
    conf.register_group(api_group)
    conf.register_opts(ALL_OPTS, api_group)


def list_opts():
    return {
        api_group: ALL_OPTS
    }



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add uuid_to_resource_class

Revision ID: 8192905fd835
Revises: e4d145e195f4
Create Date: 2017-02-24 07:00:22.344162

"""

# revision identifiers, used by Alembic.
revision = '8192905fd835'
down_revision = 'e4d145e195f4'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('resource_class',
                  sa.Column('uuid', sa.String(length=36), nullable=False))
    op.create_unique_constraint('uniq_resource_class0uuid',
                                'resource_class', ['uuid'])
    op.drop_index('uniq_container0name', table_name='resource_class')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating Container Actions via the DB API"""

from datetime import timedelta
from oslo_utils import timeutils
from oslo_utils import uuidutils

from zun.common import exception
import zun.conf
from zun.db import api as dbapi
from zun.tests.unit.db import base
from zun.tests.unit.db import utils

CONF = zun.conf.CONF


class _DBContainerActionBase(base.DbTestCase,
                             base.ModelsObjectComparatorMixin):

    IGNORED_FIELDS = [
        'id',
        'created_at',
        'updated_at',
        'details'
    ]

    def _create_action_values(self, uuid, action='create_container'):
        utils.create_test_container(context=self.context,
                                    name='cont1',
                                    uuid=uuid)

        values = {
            'action': action,
            'container_uuid': uuid,
            'request_id': self.context.request_id,
            'user_id': self.context.user_id,
            'project_id': self.context.project_id,
            'start_time': timeutils.utcnow(),
            'message': 'action-message'
        }
        return values

    def _create_event_values(self, uuid, event='do_create', extra=None):
        values = {
            'event': event,
            'container_uuid': uuid,
            'request_id': self.context.request_id,
            'start_time': timeutils.utcnow()
        }
        if extra is not None:
            values.update(extra)
        return values

    def _assertActionSaved(self, action, uuid):
        """Retrieve the action to ensure it was successfully added."""
        actions = dbapi.actions_get(self.context, uuid)
        self.assertEqual(1, len(actions))
        self._assertEqualObjects(action, actions[0])

    def _assertActionEventSaved(self, event, action_id):
        """Retrieve the event to ensure it was successfully added."""
        events = dbapi.action_events_get(self.context, action_id)
        self.assertEqual(1, len(events))
        self._assertEqualObjects(event, events[0],
                                 ['container_uuid', 'request_id'])


class DbContainerActionTestCase(_DBContainerActionBase):

    def setUp(self):
        super(DbContainerActionTestCase, self).setUp()

    def test_container_action_start(self):
        """Create a container action."""
        uuid = uuidutils.generate_uuid()
        action_values = self._create_action_values(uuid)
        action = dbapi.action_start(self.context, action_values)

        ignored_keys = self.IGNORED_FIELDS + ['finish_time']
        self._assertEqualObjects(action_values, action, ignored_keys)

        self._assertActionSaved(action, uuid)

    def test_container_actions_get_by_container(self):
        """Ensure we can get actions by UUID."""
        uuid1 = uuidutils.generate_uuid()

        expected = []

        action_values = self._create_action_values(uuid1)
        action = dbapi.action_start(self.context, action_values)
        expected.append(action)

        action_values['action'] = 'test-action'
        action = dbapi.action_start(self.context, action_values)
        expected.append(action)

        # Create an other container action.
        uuid2 = uuidutils.generate_uuid()
        action_values = self._create_action_values(uuid2, 'test-action')
        dbapi.action_start(self.context, action_values)

        actions = dbapi.actions_get(self.context, uuid1)
        self._assertEqualListsOfObjects(expected, actions)

    def test_container_action_get_by_container_and_request(self):
        """Ensure we can get an action by container UUID and request_id"""
        uuid1 = uuidutils.generate_uuid()

        action_values = self._create_action_values(uuid1)
        dbapi.action_start(self.context, action_values)
        request_id = action_values['request_id']

        # An other action using a different req id
        action_values['action'] = 'test-action'
        action_values['request_id'] = 'req-00000000-7522-4d99-7ff-111111111111'
        dbapi.action_start(self.context, action_values)

        action = dbapi.action_get_by_request_id(self.context, uuid1,
                                                request_id)
        self.assertEqual('create_container', action['action'])
        self.assertEqual(self.context.request_id, action['request_id'])

    def test_container_action_event_start(self):
        """Create a container action event."""
        uuid = uuidutils.generate_uuid()

        action_values = self._create_action_values(uuid)
        action = dbapi.action_start(self.context, action_values)

        event_values = self._create_event_values(uuid)
        event = dbapi.action_event_start(self.context, event_values)

        event_values['action_id'] = action['id']
        ignored_keys = self.IGNORED_FIELDS + ['finish_time', 'traceback',
                                              'result']
        self._assertEqualObjects(event_values, event, ignored_keys)

        self._assertActionEventSaved(event, action['id'])

    def test_container_action_event_start_without_action(self):
        uuid = uuidutils.generate_uuid()

        event_values = self._create_event_values(uuid)
        self.assertRaises(exception.ContainerActionNotFound,
                          dbapi.action_event_start, self.context, event_values)

    def test_container_action_event_finish_success(self):
        """Finish a container action event."""
        uuid = uuidutils.generate_uuid()

        action = dbapi.action_start(self.context,
                                    self._create_action_values(uuid))

        dbapi.action_event_start(self.context,
                                 self._create_event_values(uuid))

        event_values = {
            'finish_time': timeutils.utcnow() + timedelta(seconds=5),
            'result': 'Success'
        }

        event_values = self._create_event_values(uuid, extra=event_values)
        event = dbapi.action_event_finish(self.context, event_values)

        self._assertActionEventSaved(event, action['id'])
        action = dbapi.action_get_by_request_id(self.context, uuid,
                                                self.context.request_id)
        self.assertNotEqual('Error', action['message'])

    def test_container_action_event_finish_without_action(self):
        uuid = uuidutils.generate_uuid()

        event_values = {
            'finish_time': timeutils.utcnow() + timedelta(seconds=5),
            'result': 'Success'
        }
        event_values = self._create_event_values(uuid, extra=event_values)
        self.assertRaises(exception.ContainerActionNotFound,
                          dbapi.action_event_finish,
                          self.context, event_values)

    def test_container_action_events_get_in_order(self):
        """Ensure retrived action events are in order."""
        uuid1 = uuidutils.generate_uuid()

        action = dbapi.action_start(self.context,
                                    self._create_action_values(uuid1))

        extra1 = {
            'created_at': timeutils.utcnow()
        }

        extra2 = {
            'created_at': timeutils.utcnow() + timedelta(seconds=5)
        }

        event_val1 = self._create_event_values(uuid1, 'fake1', extra=extra1)
        event_val2 = self._create_event_values(uuid1, 'fake2', extra=extra1)
        event_val3 = self._create_event_values(uuid1, 'fake3', extra=extra2)

        event1 = dbapi.action_event_start(self.context, event_val1)
        event2 = dbapi.action_event_start(self.context, event_val2)
        event3 = dbapi.action_event_start(self.context, event_val3)

        events = dbapi.action_events_get(self.context, action['id'])

        self.assertEqual(3, len(events))

        self._assertEqualOrderedListOfObjects([event3, event2, event1], events,
                                              ['container_uuid', 'request_id'])



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from zun.conf import api
from zun.conf import availability_zone
from zun.conf import cinder_client
from zun.conf import cni_daemon
from zun.conf import compute
from zun.conf import container_driver
from zun.conf import database
from zun.conf import docker
from zun.conf import glance_client
from zun.conf import image_driver
from zun.conf import keystone
from zun.conf import netconf
from zun.conf import network
from zun.conf import neutron
from zun.conf import neutron_client
from zun.conf import path
from zun.conf import pci
from zun.conf import placement_client
from zun.conf import profiler
from zun.conf import quota
from zun.conf import scheduler
from zun.conf import services
from zun.conf import ssl
from zun.conf import utils
from zun.conf import volume
from zun.conf import websocket_proxy
from zun.conf import zun_client

CONF = cfg.CONF

api.register_opts(CONF)
compute.register_opts(CONF)
container_driver.register_opts(CONF)
database.register_opts(CONF)
docker.register_opts(CONF)
glance_client.register_opts(CONF)
image_driver.register_opts(CONF)
keystone.register_opts(CONF)
path.register_opts(CONF)
scheduler.register_opts(CONF)
services.register_opts(CONF)
zun_client.register_opts(CONF)
ssl.register_opts(CONF)
profiler.register_opts(CONF)
neutron_client.register_opts(CONF)
network.register_opts(CONF)
websocket_proxy.register_opts(CONF)
pci.register_opts(CONF)
quota.register_opts(CONF)
volume.register_opts(CONF)
cinder_client.register_opts(CONF)
netconf.register_opts(CONF)
availability_zone.register_opts(CONF)
utils.register_opts(CONF)
placement_client.register_opts(CONF)
cni_daemon.register_opts(CONF)
neutron.register_opts(CONF)



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from zun.pci import whitelist
from zun.tests import base


dev_dict = {
    'compute_node_id': 1,
    'address': '0000:00:0a.1',
    'product_id': '0001',
    'vendor_id': '8086',
    'status': 'available',
    'phys_function': '0000:00:0a.0',
    }


class WhitelistTestCase(base.TestCase):
    def test_whitelist(self):
        white_list = '{"product_id":"0001", "vendor_id":"8086"}'
        parsed = whitelist.Whitelist([white_list])
        self.assertEqual(1, len(parsed.specs))

    def test_whitelist_list_format(self):
        white_list = '[{"product_id":"0001", "vendor_id":"8086"},'\
                     '{"product_id":"0002", "vendor_id":"8086"}]'
        parsed = whitelist.Whitelist([white_list])
        self.assertEqual(2, len(parsed.specs))

    def test_whitelist_empty(self):
        parsed = whitelist.Whitelist()
        self.assertFalse(parsed.device_assignable(dev_dict))

    def test_whitelist_multiple(self):
        wl1 = '{"product_id":"0001", "vendor_id":"8086"}'
        wl2 = '{"product_id":"0002", "vendor_id":"8087"}'
        parsed = whitelist.Whitelist([wl1, wl2])
        self.assertEqual(2, len(parsed.specs))

    def test_device_assignable_glob(self):
        white_list = '{"address":"*:00:0a.*", "physical_network":"hr_net"}'
        parsed = whitelist.Whitelist(
            [white_list])
        self.assertTrue(parsed.device_assignable(dev_dict))

    def test_device_not_assignable_glob(self):
        white_list = '{"address":"*:00:0b.*", "physical_network":"hr_net"}'
        parsed = whitelist.Whitelist(
            [white_list])
        self.assertFalse(parsed.device_assignable(dev_dict))

    def test_device_assignable_regex(self):
        white_list = ('{"address":{"domain": ".*", "bus": ".*", '
                      '"slot": "0a", "function": "[0-1]"}, '
                      '"physical_network":"hr_net"}')
        parsed = whitelist.Whitelist(
            [white_list])
        self.assertTrue(parsed.device_assignable(dev_dict))

    def test_device_not_assignable_regex(self):
        white_list = ('{"address":{"domain": ".*", "bus": ".*", '
                      '"slot": "0a", "function": "[2-3]"}, '
                      '"physical_network":"hr_net"}')
        parsed = whitelist.Whitelist(
            [white_list])
        self.assertFalse(parsed.device_assignable(dev_dict))

    def test_device_assignable(self):
        white_list = '{"product_id":"0001", "vendor_id":"8086"}'
        parsed = whitelist.Whitelist([white_list])
        self.assertTrue(parsed.device_assignable(dev_dict))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

driver_opts = [
    cfg.StrOpt('container_driver',
               default='docker',
               help="""Defines which driver to use for controlling container.
Possible values:

* ``docker``

Services which consume this:

* ``zun-compute``

Interdependencies to other options:

* None
"""),
    cfg.StrOpt('capsule_driver',
               default='cri',
               help="""Defines which driver to use for controlling capsule.
Possible values:

* ``docker``
* ``cri``

Services which consume this:

* ``zun-compute``

Interdependencies to other options:

* None
"""),
    cfg.IntOpt('default_sleep_time', default=1,
               help='Time to sleep (in seconds) during waiting for an event.'),
    cfg.IntOpt('default_timeout', default=60 * 10,
               help='Maximum time (in seconds) to wait for an event.'),
    cfg.StrOpt('floating_cpu_set',
               default="",
               help='Define the cpusets to be excluded from pinning'),
    cfg.StrOpt('container_runtime', default='runc',
               help="""Define the runtime to create container with. Default value
in Zun is ``runc``."""),
    cfg.IntOpt('default_memory_swap',
               default=-1,
               help='The default memory swap size in MB (default is -1 '
                    'which enable unlimited swap).'),
    cfg.IntOpt('minimum_memory',
               default=4,
               help='The minimum memory size in MB allowed to set '
                    'when run/create container.'),
    cfg.IntOpt('maximum_memory',
               default=8192,
               help='The maximum memory size in MB allowed to set '
                    'when run/create container.'),
    cfg.FloatOpt('minimum_cpus',
                 default=0.1,
                 help='The minimum number of virtual cpus allowed to set '
                 'when run/create container.'),
    cfg.FloatOpt('maximum_cpus',
                 default=16.0,
                 help='The maximum number of virtual cpus allowed to set '
                 'when run/create container.'),
    cfg.IntOpt('minimum_disk',
               default=1,
               help='The minimum disk size in GB that user can set '
                    'when run/create container.'),
    cfg.IntOpt('maximum_disk',
               default=160,
               help='The maximum disk size in GB that user can set '
                    'when run/create container.'),
    cfg.IntOpt('default_memory',
               default=512,
               help='The default memory in MB a container can use '
                    '(will be used if user do not specify '
                    'container\'s memory). This value should be '
                    'in range [minimum_memory, maximum_memory].'),
    cfg.FloatOpt('default_cpu',
                 default=1.0,
                 help='The default number of cpus a container can use '
                 '(will be used if user do not specify '
                 'a container\'s cpus). This value should be '
                 'in range [minimum_cpus, maximum_cpus]'),
    cfg.IntOpt('default_disk',
               default=10,
               help='The default disk size a container can use '
                    '(will be used if user do not specify '
                    'container\'s disk). This value should be '
                    'in range [minimum_disk, maximum_disk]. Default '
                    'is 10 (GiB).')
]


ALL_OPTS = (driver_opts)


def register_opts(conf):
    conf.register_opts(ALL_OPTS)


def list_opts():
    return {"DEFAULT": ALL_OPTS}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from zun.common import consts
import zun.conf


CONF = zun.conf.CONF
LOG = logging.getLogger(__name__)


def compute_status_filter(ctxt, extra_specs):
    """Pre-filter node resource providers using ZUN_COMPUTE_STATUS_DISABLED

    The ComputeFilter filters out hosts for compute services that are
    disabled. Compute node resource providers managed by a disabled compute
    service should have the ZUN_COMPUTE_STATUS_DISABLED trait set and be
    excluded by this mandatory pre-filter.
    """
    trait_name = consts.ZUN_COMPUTE_STATUS_DISABLED
    extra_specs['trait:%s' % trait_name] = 'forbidden'
    LOG.debug('compute_status_filter request filter added forbidden '
              'trait %s', trait_name)
    return True


ALL_REQUEST_FILTERS = [
    compute_status_filter,
]


def process_reqspec(ctxt, extra_specs):
    """Process an objects.ReqestSpec before calling placement."""
    for filter in ALL_REQUEST_FILTERS:
        filter(ctxt, extra_specs)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import itertools

from zun.common import utils


_basic_keys = (
    'availability_zone',
    'binary',
    'created_at',
    'disabled',
    'disabled_reason',
    'forced_down',
    'host',
    'id',
    'last_seen_up',
    'report_count',
    'state',
    'updated_at',
)


def format_service(service):
    def transform(key, value):
        if key not in _basic_keys:
            return
        if isinstance(value, datetime.datetime):
            yield (key, utils.strtime(value))
        else:
            yield (key, value)

    return dict(
        itertools.chain.from_iterable(
            transform(k, v)for k, v in service.items()))



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import pecan
from pecan import rest

from zun.api.controllers import base
from zun.api.controllers import link
from zun.api.controllers import v1
from zun.api.controllers import versions


class Version(base.APIBase):
    """An API version representation."""

    fields = (
        'id',
        'links',
        'status',
        'max_version',
        'min_version'
    )

    @staticmethod
    def convert(id, status, max, min):
        version = Version()
        version.id = id
        version.links = [link.make_link('self', pecan.request.host_url,
                                        id, '', bookmark=True)]
        version.status = status
        version.max_version = max
        version.min_version = min
        return version


class Root(base.APIBase):

    fields = (
        'name',
        'description',
        'versions',
        'default_version',
    )

    @staticmethod
    def convert():
        root = Root()
        root.name = "OpenStack Zun API"
        root.description = ("Zun is an OpenStack project which aims to "
                            "provide containers service.")

        root.versions = [Version.convert('v1', "CURRENT",
                                         versions.CURRENT_MAX_VER,
                                         versions.BASE_VER)]
        root.default_version = Version.convert('v1', "CURRENT",
                                               versions.CURRENT_MAX_VER,
                                               versions.BASE_VER)
        return root


class RootController(rest.RestController):

    _versions = ['v1']
    """All supported API versions"""

    _default_version = 'v1'
    """The default API version"""

    v1 = v1.Controller()

    @pecan.expose('json')
    def get(self):
        # NOTE: The reason why convert() it's being called for every
        #       request is because we need to get the host url from
        #       the request object to make the links.
        return Root.convert()

    @pecan.expose()
    def _route(self, args):
        """Overrides the default routing behavior.

        It redirects the request to the default version of the zun API
        if the version number is not specified in the url.
        """

        if args[0] and args[0] not in self._versions:
            args = [self._default_version] + args
        return super(RootController, self)._route(args)



# Copyright 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import importlib
import inspect
from unittest import mock

from oslo_utils import importutils
from osprofiler import initializer as profiler_init
from osprofiler import opts as profiler_opts

from zun.common import profiler
from zun import conf
from zun.tests import base


class TestProfiler(base.TestCase):
    def test_all_public_methods_are_traced(self):
        profiler_opts.set_defaults(conf.CONF)
        self.config(enabled=True,
                    group='profiler')

        classes = [
            'zun.compute.api.API',
            'zun.compute.rpcapi.API',
        ]
        for clsname in classes:
            # give the metaclass and trace_cls() decorator a chance to patch
            # methods of the classes above
            importlib.reload(
                importutils.import_module(clsname.rsplit('.', 1)[0]))
            cls = importutils.import_class(clsname)

            for attr, obj in cls.__dict__.items():
                # only public methods are traced
                if attr.startswith('_'):
                    continue
                # only checks callables
                if not (inspect.ismethod(obj) or inspect.isfunction(obj)):
                    continue
                # osprofiler skips static methods
                if isinstance(obj, staticmethod):
                    continue

                self.assertTrue(getattr(obj, '__traced__', False), obj)

    @mock.patch.object(profiler_init, 'init_from_conf')
    def test_setup_profiler(self, mock_init):
        self.config(enabled=True,
                    group='profiler')

        profiler.setup('foo', 'localhost')

        mock_init.assert_called_once_with(conf=conf.CONF,
                                          context=mock.ANY,
                                          project="zun",
                                          service='foo',
                                          host='localhost')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import sys

import os_resource_classes as orc
from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import units
import psutil
from stevedore import driver as stevedore_driver

from zun.common import exception
from zun.common.i18n import _
from zun.common import utils
import zun.conf
from zun.container.os_capability.linux import os_capability_linux
from zun import objects
from zun.volume import driver as vol_driver


LOG = logging.getLogger(__name__)
CONF = zun.conf.CONF


# TODO(hongbin): define a list of standard traits keyed by capabilities
CAPABILITY_TRAITS_MAP = {}


def load_container_driver(container_driver=None):
    """Load a container driver module.

    Load the container driver module specified by the container_driver
    configuration option or, if supplied, the driver name supplied as an
    argument.
    :param container_driver: a container driver name to override the config opt
    :returns: a ContainerDriver instance
    """
    if not container_driver:
        container_driver = CONF.container_driver
        if not container_driver:
            LOG.error("Container driver option required, "
                      "but not specified")
            sys.exit(1)

    LOG.info("Loading container driver '%s'", container_driver)
    try:
        if container_driver.startswith('docker.driver.'):
            # case 1: (deprecated) CONF.container_driver is
            # 'docker.driver.DockerDriver'
            container_driver = 'zun.container.%s' % container_driver
            driver = importutils.import_object(container_driver)
        elif container_driver.startswith('zun.'):
            # case 2: (deprecated) CONF.container_driver is
            # 'zun.container.docker.driver.DockerDriver'
            driver = importutils.import_object(container_driver)
        else:
            # case 3: CONF.container_driver is (for example) 'docker'
            # load from entry point in this case.
            driver = stevedore_driver.DriverManager(
                "zun.container.driver",
                container_driver,
                invoke_on_load=True).driver

        if not isinstance(driver, ContainerDriver):
            raise Exception(_('Expected driver of type: %s') %
                            str(ContainerDriver))

        return driver
    except ImportError:
        LOG.exception("Unable to load the container driver")
        sys.exit(1)


def load_capsule_driver():
    driver = stevedore_driver.DriverManager(
        "zun.capsule.driver",
        CONF.capsule_driver,
        invoke_on_load=True).driver

    if not isinstance(driver, CapsuleDriver):
        raise Exception(_('Expected driver of type: %s') %
                        str(ContainerDriver))

    return driver


class BaseDriver(object):
    """Base class for driver."""

    def __init__(self):
        self.volume_drivers = {}
        for driver_name in CONF.volume.driver_list:
            driver = vol_driver.driver(driver_name)
            self.volume_drivers[driver_name] = driver
        self.cpu_allocation_ratio = CONF.compute.cpu_allocation_ratio
        self.ram_allocation_ratio = CONF.compute.ram_allocation_ratio

    def get_host_numa_topology(self):
        numa_topo_obj = objects.NUMATopology()
        os_capability_linux.LinuxHost().get_host_numa_topology(numa_topo_obj)
        return numa_topo_obj

    def get_pci_resources(self):
        return os_capability_linux.LinuxHost().get_pci_resources()

    def get_host_mem(self):
        return os_capability_linux.LinuxHost().get_host_mem()

    def get_total_disk_for_container(self):
        disk_usage = psutil.disk_usage('/')
        total_disk = disk_usage.total / 1024 ** 3
        # TODO(hongbin): deprecate reserve_disk_for_image in flavor of
        # reserved_host_disk_mb
        return (int(total_disk),
                int(total_disk * CONF.compute.reserve_disk_for_image))

    def _get_volume_driver(self, volume_mapping):
        driver_name = volume_mapping.volume_provider
        driver = self.volume_drivers.get(driver_name)
        if not driver:
            msg = _("The volume provider '%s' is not supported") % driver_name
            raise exception.ZunException(msg)

        return driver

    def attach_volume(self, context, volume_mapping):
        volume_driver = self._get_volume_driver(volume_mapping)
        volume_driver.attach(context, volume_mapping)

    def detach_volume(self, context, volume_mapping):
        volume_driver = self._get_volume_driver(volume_mapping)
        volume_driver.detach(context, volume_mapping)

    def delete_volume(self, context, volume_mapping):
        volume_driver = self._get_volume_driver(volume_mapping)
        volume_driver.delete(context, volume_mapping)

    def is_volume_available(self, context, volume_mapping):
        volume_driver = self._get_volume_driver(volume_mapping)
        return volume_driver.is_volume_available(context, volume_mapping)

    def is_volume_deleted(self, context, volume_mapping):
        volume_driver = self._get_volume_driver(volume_mapping)
        return volume_driver.is_volume_deleted(context, volume_mapping)

    def get_available_nodes(self):
        return [CONF.host]

    def node_support_disk_quota(self):
        return False

    def get_host_default_base_size(self):
        return None

    def get_available_resources(self):
        """Retrieve resource information.

        This method is called when zun-compute launches, and
        as part of a periodic task that records the results in the DB.

        :returns: dictionary containing resource info
        """
        data = {}

        numa_topo_obj = self.get_host_numa_topology()
        data['numa_topology'] = numa_topo_obj
        meminfo = self.get_host_mem()
        (mem_total, mem_free, mem_ava, mem_used) = meminfo
        data['mem_total'] = mem_total // units.Ki
        data['mem_free'] = mem_free // units.Ki
        data['mem_available'] = mem_ava // units.Ki
        data['mem_used'] = mem_used // units.Ki
        disk_total, disk_reserved = self.get_total_disk_for_container()
        data['disk_total'] = disk_total - disk_reserved
        disk_quota_supported = self.node_support_disk_quota()
        data['disk_quota_supported'] = disk_quota_supported
        data['enable_cpu_pinning'] = CONF.compute.enable_cpu_pinning

        return data

    def node_is_available(self, nodename):
        """Return whether this compute service manages a particular node."""
        if nodename in self.get_available_nodes():
            return True
        return False

    def update_provider_tree(self, provider_tree, nodename):
        """Update a ProviderTree object with current resource provider,
        inventory information.
        :param zun.compute.provider_tree.ProviderTree provider_tree:
            A zun.compute.provider_tree.ProviderTree object representing all
            the providers in the tree associated with the compute node, and any
            sharing providers (those with the ``MISC_SHARES_VIA_AGGREGATE``
            trait) associated via aggregate with any of those providers (but
            not *their* tree- or aggregate-associated providers), as currently
            known by placement.
        :param nodename:
            String name of the compute node (i.e. ComputeNode.hostname)
            for which the caller is requesting updated provider information.
        """
        def _get_local_gb_info():
            return self.get_total_disk_for_container()[0]

        def _get_memory_mb_total():
            mem_total, mem_free, mem_ava, mem_used = self.get_host_mem()
            return mem_total // units.Ki

        def _get_vcpu_total():
            return psutil.cpu_count()

        disk_gb = _get_local_gb_info()
        memory_mb = _get_memory_mb_total()
        vcpus = _get_vcpu_total()

        # NOTE(yikun): If the inv record does not exists, the allocation_ratio
        # will use the CONF.xxx_allocation_ratio value if xxx_allocation_ratio
        # is set, and fallback to use the initial_xxx_allocation_ratio
        # otherwise.
        inv = provider_tree.data(nodename).inventory
        ratios = self._get_allocation_ratios(inv)
        self.cpu_allocation_ratio = ratios[orc.VCPU]
        self.ram_allocation_ratio = ratios[orc.MEMORY_MB]
        result = {
            orc.VCPU: {
                'total': vcpus,
                'min_unit': 1,
                'max_unit': vcpus,
                'step_size': 1,
                'allocation_ratio': ratios[orc.VCPU],
                # TODO(hongbin): handle the case that the zun's reserved value
                # override the nova's one
                'reserved': CONF.compute.reserved_host_cpus,
            },
            orc.MEMORY_MB: {
                'total': memory_mb,
                'min_unit': 1,
                'max_unit': memory_mb,
                'step_size': 1,
                'allocation_ratio': ratios[orc.MEMORY_MB],
                # TODO(hongbin): handle the case that the zun's reserved value
                # override the nova's one
                'reserved': CONF.compute.reserved_host_memory_mb,
            },
        }

        # If a sharing DISK_GB provider exists in the provider tree, then our
        # storage is shared, and we should not report the DISK_GB inventory in
        # the compute node provider.
        # TODO(efried): Reinstate non-reporting of shared resource by the
        # compute RP once the issues from bug #1784020 have been resolved.
        if provider_tree.has_sharing_provider(orc.DISK_GB):
            LOG.debug('Ignoring sharing provider - see bug #1784020')
        result[orc.DISK_GB] = {
            'total': disk_gb,
            'min_unit': 1,
            'max_unit': disk_gb,
            'step_size': 1,
            'allocation_ratio': ratios[orc.DISK_GB],
            # TODO(hongbin): handle the case that the zun's reserved value
            # override the nova's one
            'reserved': self._get_reserved_host_disk_gb_from_config(),
        }

        provider_tree.update_inventory(nodename, result)

        # Now that we updated the ProviderTree, we want to store it locally
        # so that spawn() or other methods can access it thru a getter
        self.provider_tree = copy.deepcopy(provider_tree)

    def get_cpu_allocation_ratio(self):
        return self.cpu_allocation_ratio

    def get_ram_allocation_ratio(self):
        return self.ram_allocation_ratio

    @staticmethod
    def _get_allocation_ratios(inventory):
        """Get the cpu/ram/disk allocation ratios for the given inventory.

        This utility method is used to get the inventory allocation ratio
        for VCPU, MEMORY_MB and DISK_GB resource classes based on the following
        precedence:

        * Use ``[DEFAULT]/*_allocation_ratio`` if set - this overrides
          everything including externally set allocation ratios on the
          inventory via the placement API
        * Use ``[DEFAULT]/initial_*_allocation_ratio`` if a value does not
          exist for a given resource class in the ``inventory`` dict
        * Use what is already in the ``inventory`` dict for the allocation
          ratio if the above conditions are false

        :param inventory: dict, keyed by resource class, of inventory
                          information.
        :returns: Return a dict, keyed by resource class, of allocation ratio
        """
        keys = {'cpu': orc.VCPU,
                'ram': orc.MEMORY_MB,
                'disk': orc.DISK_GB}
        result = {}
        for res, rc in keys.items():
            attr = '%s_allocation_ratio' % res
            conf_ratio = getattr(CONF.compute, attr)
            if conf_ratio:
                result[rc] = conf_ratio
            elif rc not in inventory:
                result[rc] = getattr(CONF.compute, 'initial_%s' % attr)
            else:
                result[rc] = inventory[rc]['allocation_ratio']
        return result

    @staticmethod
    def _get_reserved_host_disk_gb_from_config():
        return utils.convert_mb_to_ceil_gb(CONF.compute.reserved_host_disk_mb)

    def capabilities_as_traits(self):
        """Returns this driver's capabilities dict where the keys are traits

        Traits can only be standard compute capabilities traits from
        the os-traits library.

        :returns: dict, keyed by trait, of this driver's capabilities where the
            values are booleans indicating if the driver supports the trait

        """
        traits = {}
        for capability, supported in self.capabilities.items():
            if capability in CAPABILITY_TRAITS_MAP:
                traits[CAPABILITY_TRAITS_MAP[capability]] = supported

        return traits


class ContainerDriver(object):
    """Interface for container driver."""

    def create(self, context, container, **kwargs):
        """Create a container."""
        raise NotImplementedError()

    def commit(self, context, container, repository, tag):
        """Commit a container."""
        raise NotImplementedError()

    def delete(self, context, container, force):
        """Delete a container."""
        raise NotImplementedError()

    def list(self, context):
        """List all containers."""
        raise NotImplementedError()

    def update_containers_states(self, context, containers, manager):
        """Update containers states."""
        raise NotImplementedError()

    def show(self, context, container):
        """Show the details of a container."""
        raise NotImplementedError()

    def reboot(self, context, container):
        """Reboot a container."""
        raise NotImplementedError()

    def stop(self, context, container):
        """Stop a container."""
        raise NotImplementedError()

    def start(self, context, container):
        """Start a container."""
        raise NotImplementedError()

    def pause(self, context, container):
        """Pause a container."""
        raise NotImplementedError()

    def unpause(self, context, container):
        """Unpause a container."""
        raise NotImplementedError()

    def show_logs(self, context, container, stdout=True, stderr=True,
                  timestamps=False, tail='all', since=None):
        """Show logs of a container."""
        raise NotImplementedError()

    def execute_create(self, context, container, command, **kwargs):
        """Create an execute instance for running a command."""
        raise NotImplementedError()

    def execute_run(self, exec_id):
        """Run the command specified by an execute instance."""
        raise NotImplementedError()

    def execute_resize(self, exec_id, height, width):
        """Resizes the tty session used by the exec."""
        raise NotImplementedError()

    def kill(self, context, container, signal):
        """Kill a container with specified signal."""
        raise NotImplementedError()

    def get_websocket_url(self, context, container):
        """Get websocket url of a container."""
        raise NotImplementedError()

    def resize(self, context, container, height, weight):
        """Resize tty of a container."""
        raise NotImplementedError()

    def top(self, context, container, ps_args):
        """Display the running processes inside the container."""
        raise NotImplementedError()

    def get_archive(self, context, container, path):
        """Copy resource from a container."""
        raise NotImplementedError()

    def put_archive(self, context, container, path, data):
        """Copy resource to a container."""
        raise NotImplementedError()

    def stats(self, context, container):
        """Display stats of the container."""
        raise NotImplementedError()

    def update(self, context, container):
        """Update a container."""
        raise NotImplementedError()

    def add_security_group(self, context, container, security_group, **kwargs):
        raise NotImplementedError()

    def remove_security_group(self, context, container, security_group,
                              **kwargs):
        raise NotImplementedError()

    def network_detach(self, context, container, network):
        raise NotImplementedError()

    def network_attach(self, context, container, requested_network):
        raise NotImplementedError()

    def create_network(self, context, network):
        raise NotImplementedError()

    def delete_network(self, context, network):
        raise NotImplementedError()

    def pull_image(self, context, repo, tag, **kwargs):
        raise NotImplementedError()

    def search_image(self, context, repo, tag, driver_name, exact_match):
        raise NotImplementedError()

    def create_image(self, context, image_name, image_driver):
        raise NotImplementedError()

    def upload_image_data(self, context, image, image_tag, image_data,
                          image_driver):
        raise NotImplementedError()

    def delete_committed_image(self, context, img_id, image_driver):
        raise NotImplementedError()

    def delete_image(self, context, img_id, image_driver):
        raise NotImplementedError()


class CapsuleDriver(object):
    """Interface for container driver."""

    def create_capsule(self, context, capsule, **kwargs):
        raise NotImplementedError()

    def delete_capsule(self, context, capsule, **kwargs):
        raise NotImplementedError()



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Add container host operating system info

Revision ID: 04ba87af76bb
Revises: 8c3d80e18eb5
Create Date: 2017-04-26 17:37:09.544598

"""

# revision identifiers, used by Alembic.
revision = '04ba87af76bb'
down_revision = '8c3d80e18eb5'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('compute_node',
                  sa.Column('architecture', sa.String(32), nullable=True))
    op.add_column('compute_node',
                  sa.Column('os_type', sa.String(32), nullable=True))
    op.add_column('compute_node',
                  sa.Column('os', sa.String(64), nullable=True))
    op.add_column('compute_node',
                  sa.Column('kernel_version',
                            sa.String(128), nullable=True))



# Copyright 2017 ARM Holdings
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add auto heal to container

Revision ID: 372433c0afd2
Revises: d0c606fdec3c
Create Date: 2018-03-12 14:18:44.545951

"""

# revision identifiers, used by Alembic.
revision = '372433c0afd2'
down_revision = 'd0c606fdec3c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('container', schema=None) as batch_op:
        batch_op.add_column(sa.Column('auto_heal',
                                      sa.Boolean(), nullable=True))
    # ### end Alembic commands ###



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zun.common.docker_image import regexp
from zun.tests import base


class TestRegexp(base.BaseTestCase):
    def test_generated_regexps(self):
        alphaNumericRegexp = r'[a-z0-9]+'
        separatorRegexp = r'(?:[._]|__|[-]*)'
        nameComponentRegexp = r'[a-z0-9]+(?:(?:(?:[._]|__|[-]*)[a-z0-9]+)+)?'
        hostnameComponentRegexp = r'(?:[a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9-]*'\
                                  r'[a-zA-Z0-9])'
        hostnameRegexp = r'(?:[a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9-]*'\
                         r'[a-zA-Z0-9])'\
                         r'(?:(?:\.(?:[a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9-]*'\
                         r'[a-zA-Z0-9]))+)?(?::[0-9]+)?'
        TagRegexp = r'[\w][\w.-]{0,127}'
        anchoredTagRegexp = r'^[\w][\w.-]{0,127}$'
        NameRegexp = r'(?:(?:[a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9-]*[a-zA-Z0-9])'\
                     r'(?:(?:\.(?:[a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9-]*'\
                     r'[a-zA-Z0-9]))+)?(?::[0-9]+)?/)?[a-z0-9]+(?:(?:'\
                     r'(?:[._]|__|[-]*)[a-z0-9]+)+)?(?:(?:/[a-z0-9]+'\
                     r'(?:(?:(?:[._]|__|[-]*)[a-z0-9]+)+)?)+)?'
        anchoredNameRegexp = r'^(?:((?:[a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9-]*'\
                             r'[a-zA-Z0-9])(?:(?:\.(?:[a-zA-Z0-9]|[a-zA-Z0-9]'\
                             r'[a-zA-Z0-9-]*[a-zA-Z0-9]))+)?(?::[0-9]+)?)/)?'\
                             r'([a-z0-9]+(?:(?:(?:[._]|__|[-]*)[a-z0-9]+)+)?'\
                             r'(?:(?:/[a-z0-9]+(?:(?:(?:[._]|__|[-]*)[a-z0-9]'\
                             r'+)+)?)+)?)$'
        ReferenceRegexp = r'^((?:(?:[a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9-]*'\
                          r'[a-zA-Z0-9])(?:(?:\.(?:[a-zA-Z0-9]|[a-zA-Z0-9]'\
                          r'[a-zA-Z0-9-]*[a-zA-Z0-9]))+)?(?::[0-9]+)?/)?'\
                          r'[a-z0-9]+(?:(?:(?:[._]|__|[-]*)[a-z0-9]+)+)?'\
                          r'(?:(?:/[a-z0-9]+(?:(?:(?:[._]|__|[-]*)[a-z0-9]+)'\
                          r'+)?)+)?)(?::([\w][\w.-]{0,127}))?(?:@'\
                          r'([a-zA-Z0-9-_+.]+:[a-fA-F0-9]+))?$'

        ImageRegexps = regexp.ImageRegexps
        self.assertEqual(alphaNumericRegexp,
                         ImageRegexps.ALPHA_NUMERIC_REGEXP.pattern)
        self.assertEqual(separatorRegexp,
                         ImageRegexps.SEPARATOR_REGEXP.pattern)
        self.assertEqual(nameComponentRegexp,
                         ImageRegexps.NAME_COMPONENT_REGEXP.pattern)
        self.assertEqual(hostnameComponentRegexp,
                         ImageRegexps.HOSTNAME_COMPONENT_REGEXP.pattern)
        self.assertEqual(hostnameRegexp,
                         ImageRegexps.HOSTNAME_REGEXP.pattern)
        self.assertEqual(TagRegexp, ImageRegexps.TAG_REGEXP.pattern)
        self.assertEqual(anchoredTagRegexp,
                         ImageRegexps.ANCHORED_TAG_REGEXP.pattern)
        self.assertEqual(NameRegexp, ImageRegexps.NAME_REGEXP.pattern)
        self.assertEqual(anchoredNameRegexp,
                         ImageRegexps.ANCHORED_NAME_REGEXP.pattern)
        self.assertEqual(ReferenceRegexp,
                         ImageRegexps.REFERENCE_REGEXP.pattern)



# Copyright 2014 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# It's based on oslo.i18n usage in OpenStack Keystone project and
# recommendations from
# https://docs.openstack.org/oslo.i18n/latest/user/usage.html

import oslo_i18n


_translators = oslo_i18n.TranslatorFactory(domain='zun')

# The primary translation function using the well-known name "_"
_ = _translators.primary



# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg


utils_opts = [
    cfg.StrOpt('rootwrap_config',
               default="/etc/zun/rootwrap.conf",
               help='Path to the rootwrap configuration file to use for '
                    'running commands as root.'),
    cfg.StrOpt('auth_encryption_key',
               secret=True,
               default='notgood but just long enough i t',
               help='Key used to encrypt authentication info in the '
                    'database. Length of this key must be 32 characters.'),
]


def register_opts(conf):
    conf.register_opts(utils_opts)


def list_opts():
    return {
        "DEFAULT": utils_opts
    }



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg


neutron_group = cfg.OptGroup(name='neutron', title='Options for neutron')

neutron_opts = [
    cfg.StrOpt('ovs_bridge',
               default='br-int',
               help="""
Default name for the Open vSwitch integration bridge.

Specifies the name of an integration bridge interface used by OpenvSwitch.
This option is only used if Neutron does not specify the OVS bridge name in
port binding responses.
"""),
]

ALL_OPTS = (neutron_opts)


def register_opts(conf):
    conf.register_group(neutron_group)
    conf.register_opts(ALL_OPTS, neutron_group)


def list_opts():
    return {neutron_group: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools

from zun.common.policies import availability_zone
from zun.common.policies import base
from zun.common.policies import capsule
from zun.common.policies import container
from zun.common.policies import container_action
from zun.common.policies import host
from zun.common.policies import image
from zun.common.policies import network
from zun.common.policies import quota
from zun.common.policies import quota_class
from zun.common.policies import registry
from zun.common.policies import zun_service


def list_rules():
    return itertools.chain(
        base.list_rules(),
        container.list_rules(),
        image.list_rules(),
        zun_service.list_rules(),
        host.list_rules(),
        capsule.list_rules(),
        network.list_rules(),
        container_action.list_rules(),
        availability_zone.list_rules(),
        quota.list_rules(),
        quota_class.list_rules(),
        registry.list_rules(),
    )



# Copyright (c) 2018 NEC, Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pecan

from zun.api.controllers import base
from zun.api.controllers.v1 import collection
from zun.api.controllers.v1.views import availability_zone_view as view
from zun.api import utils as api_utils
from zun.common import exception
from zun.common import policy
import zun.conf
from zun import objects


CONF = zun.conf.CONF


def check_policy_on_availability_zones(availability_zone, action):
    context = pecan.request.context
    policy.enforce(context, action, availability_zone, action=action)


class AvailabilityZoneCollection(collection.Collection):
    """API representation of a collection of availability zones."""

    fields = {
        'availability_zones',
        'next'
    }

    """A list containing availability zone objects"""

    def __init__(self, **kwargs):
        super(AvailabilityZoneCollection, self).__init__(**kwargs)
        self._type = 'availability_zones'

    @staticmethod
    def convert_with_links(zones, limit, url=None,
                           expand=False, **kwargs):
        collection = AvailabilityZoneCollection()
        collection.availability_zones = [
            view.format_a_zone(url, p) for p in zones]
        collection.next = collection.get_next(limit, url=url, **kwargs)
        return collection


class AvailabilityZoneController(base.Controller):
    """Availability Zone info controller"""

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def get_all(self, **kwargs):
        """Retrieve a list of availability zones"""

        context = pecan.request.context
        context.all_projects = True

        policy.enforce(context, "availability_zones:get_all",
                       action="availability_zones:get_all")
        return self._get_host_collection(**kwargs)

    def _get_host_collection(self, **kwargs):
        context = pecan.request.context
        limit = api_utils.validate_limit(kwargs.get('limit'))

        sort_dir = api_utils.validate_sort_dir(kwargs.get('sort_dir', 'asc'))
        sort_key = kwargs.get('sort_key', 'availability_zone')
        expand = kwargs.get('expand')
        marker_obj = None
        resource_url = kwargs.get('resource_url')
        marker = kwargs.get('marker')
        if marker:
            marker_obj = objects.ZunService.get_by_uuid(context, marker)
        services = objects.ZunService.list(context,
                                           limit,
                                           marker_obj,
                                           sort_key,
                                           sort_dir)
        zones = {}
        for service in services:
            zones[service.availability_zone] = service
        return AvailabilityZoneCollection.convert_with_links(zones.values(),
                                                             limit,
                                                             url=resource_url,
                                                             expand=expand,
                                                             sort_key=sort_key,
                                                             sort_dir=sort_dir)



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock
from unittest.mock import patch

from oslo_utils import uuidutils

from zun import objects
from zun.objects import numa
from zun.tests.unit.api import base as api_base
from zun.tests.unit.db import utils


class TestHostController(api_base.FunctionalTest):

    @mock.patch('zun.common.policy.enforce')
    @patch('zun.objects.ComputeNode.list')
    def test_get_all_hosts(self, mock_host_list, mock_policy):
        mock_policy.return_value = True
        test_host = utils.get_test_compute_node()
        numat = numa.NUMATopology._from_dict(test_host['numa_topology'])
        test_host['numa_topology'] = numat
        hosts = [objects.ComputeNode(self.context, **test_host)]
        mock_host_list.return_value = hosts

        response = self.get('/v1/hosts')

        mock_host_list.assert_called_once_with(mock.ANY,
                                               1000, None, 'hostname', 'asc',
                                               filters=None)
        self.assertEqual(200, response.status_int)
        actual_hosts = response.json['hosts']
        self.assertEqual(1, len(actual_hosts))
        self.assertEqual(test_host['uuid'],
                         actual_hosts[0].get('uuid'))

    @mock.patch('zun.common.policy.enforce')
    @patch('zun.objects.ComputeNode.list')
    def test_get_all_hosts_with_pagination_marker(self, mock_host_list,
                                                  mock_policy):
        mock_policy.return_value = True
        host_list = []
        for id_ in range(4):
            test_host = utils.create_test_compute_node(
                context=self.context,
                uuid=uuidutils.generate_uuid(),
                rp_uuid=uuidutils.generate_uuid())
            numat = numa.NUMATopology._from_dict(test_host['numa_topology'])
            test_host['numa_topology'] = numat
            host = objects.ComputeNode(self.context, **test_host)
            host_list.append(host)
        mock_host_list.return_value = host_list[-1:]
        response = self.get('/v1/hosts?limit=3&marker=%s'
                            % host_list[2].uuid)

        self.assertEqual(200, response.status_int)
        actual_hosts = response.json['hosts']
        self.assertEqual(1, len(actual_hosts))
        self.assertEqual(host_list[-1].uuid,
                         actual_hosts[0].get('uuid'))

    @mock.patch('zun.common.policy.enforce')
    @patch('zun.objects.ComputeNode.get_by_uuid')
    def test_get_one_host(self, mock_get_by_uuid, mock_policy):
        mock_policy.return_value = True
        test_host = utils.get_test_compute_node()
        numat = numa.NUMATopology._from_dict(test_host['numa_topology'])
        test_host['numa_topology'] = numat
        test_host_obj = objects.ComputeNode(self.context, **test_host)
        mock_get_by_uuid.return_value = test_host_obj
        response = self.get('/v1/hosts/%s' % test_host['uuid'])
        mock_get_by_uuid.assert_called_once_with(
            mock.ANY,
            test_host['uuid'])
        self.assertEqual(200, response.status_int)
        self.assertEqual(test_host['uuid'],
                         response.json['uuid'])


class TestHostEnforcement(api_base.FunctionalTest):

    def _common_policy_check(self, rule, func, *arg, **kwarg):
        self.policy.set_rules({rule: 'project_id:non_fake'})
        response = func(*arg, **kwarg)
        self.assertEqual(403, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(
            "Policy doesn't allow %s to be performed." % rule,
            response.json['errors'][0]['detail'])

    def test_policy_disallow_get_all(self):
        self._common_policy_check(
            'host:get_all', self.get_json, '/hosts',
            expect_errors=True)

    def test_pollicy_disallow_get_one(self):
        self._common_policy_check(
            'host:get', self.get_json, '/hosts/%s' % '12345678',
            expect_errors=True)



# Copyright 2017 ARM Holdings
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""create capsule table

Revision ID: a251f1f61217
Revises: 75315e219cfb
Create Date: 2017-06-20 17:12:56.105277

"""

# revision identifiers, used by Alembic.
revision = 'a251f1f61217'
down_revision = '75315e219cfb'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from zun.db.sqlalchemy import models


def upgrade():
    op.create_table(
        'capsule',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('capsule_version', sa.String(length=255), nullable=True),
        sa.Column('kind', sa.String(length=36), nullable=True),
        sa.Column('project_id', sa.String(length=255), nullable=True),
        sa.Column('user_id', sa.String(length=255), nullable=True),
        sa.Column('restart_policy', sa.String(length=255), nullable=True),
        sa.Column('host_selector', sa.String(length=255), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False),
        sa.Column('status', sa.String(length=255), nullable=True),
        sa.Column('status_reason', sa.Text(), nullable=True),
        sa.Column('message', models.JSONEncodedDict(), nullable=True),
        sa.Column('spec', models.JSONEncodedDict(), nullable=True),
        sa.Column('cpu', sa.Float(), nullable=True),
        sa.Column('memory', sa.String(length=255), nullable=True),
        sa.Column('meta_name', sa.String(length=255), nullable=True),
        sa.Column('meta_labels', models.JSONEncodedList(), nullable=True),
        sa.Column('containers_uuids', models.JSONEncodedList(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
    )



#    Copyright 2017 Linaro Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

wsproxy_group = cfg.OptGroup("websocket_proxy",
                             title="Websocket Proxy Group",
                             help="""
Users use the websocket proxy to connect to containers, instead of
connecting to containers directly, hence protects the socket daemon.
""")

wsproxy_opts = [
    cfg.URIOpt('base_url',
               default='ws://$wsproxy_host:$wsproxy_port/',
               help="""
The URL an end user would use to connect to the ``zun-wsproxy`` service.

The ``zun-wsproxy`` service is called with this token enriched URL
and establishes the connection to the proper instance.

Related options:

* The IP address must be the same as the address to which the
  ``zun-wsproxy`` service is listening (see option ``wsproxy_host``
  in this section).
* The port must be the same as ``wsproxy_port``in this section.
    """),
    cfg.StrOpt('wsproxy_host',
               default='$my_ip',
               help="""
The IP address which is used by the ``zun-wsproxy`` service to listen
for incoming requests.

The ``zun-wsproxy`` service listens on this IP address for incoming
connection requests.

Related options:

* Ensure that this is the same IP address which is defined in the option
  ``base_url`` of this section or use ``0.0.0.0`` to listen on all addresses.
"""),
    cfg.PortOpt('wsproxy_port',
                default=6784,
                help="""
The port number which is used by the ``zun-wsproxy`` service to listen
for incoming requests.

The ``zun-wsproxy`` service listens on this port number for incoming
connection requests.

Related options:

* Ensure that this is the same port number as that defined in the option
  ``base_url`` of this section.
"""),
    cfg.ListOpt('allowed_origins',
                default=[],
                help="""
Adds list of allowed origins to the console websocket proxy to allow
connections from other origin hostnames.
Websocket proxy matches the host header with the origin header to
prevent cross-site requests. This list specifies if any there are
values other than host are allowed in the origin header.

Possible values:

* A list where each element is an allowed origin hostnames, else an empty list
"""),
    cfg.BoolOpt('ssl_only',
                help="Disallow non-encrypted connections."),
    cfg.StrOpt('cert',
               default='',
               help="Path to SSL certificate file."),
    cfg.StrOpt('key',
               default='',
               help="SSL key file (if separate from cert)."),
]

ALL_OPTS = (wsproxy_opts)


def register_opts(conf):
    conf.register_group(wsproxy_group)
    conf.register_opts(wsproxy_opts, group=wsproxy_group)


def list_opts():
    return {wsproxy_group: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add container_actions table

Revision ID: cf46a28f46bc
Revises: f046346d1d87
Create Date: 2017-12-01 10:47:00.192171

"""

# revision identifiers, used by Alembic.
revision = 'cf46a28f46bc'
down_revision = 'd2affd5b4172'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'container_actions',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('action', sa.String(length=255), nullable=True),
        sa.Column('container_uuid', sa.String(length=36), nullable=False),
        sa.Column('request_id', sa.String(length=255), nullable=True),
        sa.Column('user_id', sa.String(length=255), nullable=True),
        sa.Column('project_id', sa.String(length=255), nullable=True),
        sa.Column('start_time', sa.DateTime(), nullable=True),
        sa.Column('finish_time', sa.DateTime(), nullable=True),
        sa.Column('message', sa.String(length=255), nullable=True),
        sa.Index('container_uuid_idx', 'container_uuid'),
        sa.Index('request_id_idx', 'request_id'),
        sa.ForeignKeyConstraint(['container_uuid'], ['container.uuid'], ),
        sa.PrimaryKeyConstraint('id')
    )



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add_restart_policy_column

Revision ID: bbcfa910a8a5
Revises: 531e4a890480
Create Date: 2017-01-10 15:10:02.746131

"""

# revision identifiers, used by Alembic.
revision = 'bbcfa910a8a5'
down_revision = '531e4a890480'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

import zun


def upgrade():
    op.add_column('container',
                  sa.Column('restart_policy',
                            zun.db.sqlalchemy.models.JSONEncodedDict(),
                            nullable=True))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import shlex
import sys

import os_vif
from oslo_privsep import priv_context

from zun.cni.daemon import service
from zun.common import config
from zun.common import service as zun_service
from zun.common import utils


def main():
    priv_context.init(root_helper=shlex.split(utils.get_root_helper()))
    zun_service.prepare_service(sys.argv)
    config.parse_args(sys.argv)

    # Initialize o.vo registry.
    os_vif.initialize()

    service.CNIDaemonServiceManager().run()


if __name__ == '__main__':
    sys.exit(main())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_versionedobjects import fields

from zun.common import exception
from zun.db import api as dbapi
from zun.objects import base


LOG = logging.getLogger(__name__)


@base.ZunObjectRegistry.register
class Volume(base.ZunPersistentObject, base.ZunObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.IntegerField(),
        'uuid': fields.UUIDField(nullable=False),
        'project_id': fields.StringField(nullable=True),
        'user_id': fields.StringField(nullable=True),
        'cinder_volume_id': fields.UUIDField(nullable=True),
        'volume_provider': fields.StringField(nullable=False),
        'connection_info': fields.SensitiveStringField(nullable=True),
        'auto_remove': fields.BooleanField(nullable=True),
        'host': fields.StringField(nullable=True),
        'contents': fields.SensitiveStringField(nullable=True),
    }

    @staticmethod
    def _from_db_object(volume, db_volume):
        """Converts a database entity to a formal object."""
        for field in volume.fields:
            setattr(volume, field, db_volume[field])

        volume.obj_reset_changes()
        return volume

    @staticmethod
    def _from_db_object_list(db_objects, cls, context):
        """Converts a list of database entities to a list of formal objects."""
        return [Volume._from_db_object(cls(context), obj)
                for obj in db_objects]

    @base.remotable_classmethod
    def get_by_id(cls, context, vol_id):
        """Find a volume mapping based on uuid.

        :param uuid: the uuid of a volume mapping.
        :param context: Security context
        :returns: a :class:`Volume` object.
        """
        db_volume = dbapi.get_volume_by_id(context, vol_id)
        volume = Volume._from_db_object(cls(context), db_volume)
        return volume

    @base.remotable
    def create(self, context):
        """Create a Volume record in the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object.

        """
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason='already created')

        values = self.obj_get_changes()
        db_volume = dbapi.create_volume(context, values)
        self._from_db_object(self, db_volume)

    @base.remotable
    def save(self, context=None):
        """Save updates to this Volume.

        Updates will be made column by column based on the result
        of self.what_changed().

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object.
        """
        updates = self.obj_get_changes()
        if 'container' in updates:
            raise exception.ObjectActionError(action='save',
                                              reason='container changed')
        updates.pop('id', None)
        dbapi.update_volume(context, self.uuid, updates)
        self.obj_reset_changes()

    @base.remotable
    def refresh(self, context=None):
        """Loads updates for this Volume.

        Loads a volume with the same id from the database and
        checks for updated attributes. Updates are applied from
        the loaded volume column by column, if there are any updates.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object.
        """
        current = self.__class__.get_by_id(self._context, self.id)
        for field in self.fields:
            if not self.obj_attr_is_set(field):
                continue
            if getattr(self, field) != getattr(current, field):
                setattr(self, field, getattr(current, field))
        self.obj_reset_changes()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
from oslo_policy import _parser
from oslo_policy import opts as policy_opts

from zun.common import policy as zun_policy
import zun.conf


CONF = zun.conf.CONF


class PolicyFixture(fixtures.Fixture):

    def _setUp(self):
        CONF(args=[], project='zun')
        policy_opts.set_defaults(CONF)
        zun_policy._ENFORCER = None
        self.addCleanup(zun_policy.init().clear)

    def set_rules(self, rules):
        policy = zun_policy._ENFORCER
        policy.set_rules({k: _parser.parse_rule(v)
                          for k, v in rules.items()})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Fakes relating to the `requests` module."""

import requests


class FakeResponse(requests.Response):
    def __init__(self, status_code, content=None, headers=None):
        """A requests.Response that can be used as a mock return_value.

        A key feature is that the instance will evaluate to True or False like
        a real Response, based on the status_code.

        Properties like ok, status_code, text, and content, and methods like
        json(), work as expected based on the inputs.

        :param status_code: Integer HTTP response code (200, 404, etc.)
        :param content: String supplying the payload content of the response.
                        Using a json-encoded string will make the json() method
                        behave as expected.
        :param headers: Dict of HTTP header values to set.
        """
        super(FakeResponse, self).__init__()
        self.status_code = status_code
        if content:
            self._content = content.encode('utf-8')
            self.encoding = 'utf-8'
        if headers:
            self.headers = headers



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zun.api.controllers.v1.schemas import parameter_types

query_param_enable = {
    'type': 'object',
    'properties': {
        'host': parameter_types.hostname,
        'binary': {
            'type': 'string', 'minLength': 1, 'maxLength': 255,
        },
    },
    'additionalProperties': False
}

query_param_disable = {
    'type': 'object',
    'properties': {
        'host': parameter_types.hostname,
        'binary': {
            'type': 'string', 'minLength': 1, 'maxLength': 255,
        },
        'disabled_reason': {
            'type': 'string', 'minLength': 1, 'maxLength': 255,
        },
    },
    'additionalProperties': False
}

query_param_force_down = {
    'type': 'object',
    'properties': {
        'host': parameter_types.hostname,
        'binary': {
            'type': 'string', 'minLength': 1, 'maxLength': 255,
        },
        'forced_down': parameter_types.boolean
    },
    'additionalProperties': False
}
